"""
This file provides code for image I/O operations, including multithreaded settings
"""
from __future__ import annotations

import copy
import enum
import json
from typing import Any

import napari
import numcodecs
import numpy as np
import cvpl_tools.im.ndblock as cvpl_ndblock
from cvpl_tools.im.ndblock import NDBlock
import dask.array as da
import shutil
import os
from cvpl_tools.ome_zarr.napari.zarr_viewer import add_ome_zarr_array_from_path


def ensure_dir_exists(dir_path, remove_if_already_exists):
    """
    If a directory does not exist, make a new directory with the name.
    This assumes the parent directory must exists; otherwise a path not
    found error will be thrown.
    Args:
        dir_path: The path of folder
        remove_if_already_exists: if True and the folder already exists, then remove it and make a new one.
    """
    if os.path.exists(dir_path):
        if remove_if_already_exists:
            shutil.rmtree(dir_path)
            os.mkdir(dir_path)
    else:
        os.mkdir(dir_path)


class ImageFormat(enum.Enum):
    NUMPY = 0
    DASK_ARRAY = 1
    NDBLOCK = 2


def chunksize_to_str(chunksize: tuple[int, ...]):
    return ','.join(str(s) for s in chunksize)


def str_to_chunksize(chunksize_str: str):
    return tuple(int(s) for s in chunksize_str.split(','))


def save(file: str,
         im,
         storage_options: dict = None):
    """Save an image object into given path

    Supported im object types:
    - np.ndarray
    - dask.Array
    - cvpl_tools.im.ndblock.NDBlock

    Storage options
        preferred_chunksize (tuple[int, ...]) = None
            chunk sizes to save as; will rechunk if different from current size; only applies to dask arrays.
        multiscale (int) = 0
            The number of downsample layers for save ome-zarr; only applies if the image is a dask image
        compressor = None
            The compressor to use to compress array or chunks

    Args:
        file: The full/relative path to the directory to be saved to
        im: Object to be saved
        storage_options: Specifies options in saving method and saved file format
    """
    if isinstance(im, np.ndarray):
        old_chunksize = im.shape
        fmt = ImageFormat.NUMPY
    elif isinstance(im, da.Array):
        old_chunksize = im.chunksize
        fmt = ImageFormat.DASK_ARRAY
    elif isinstance(im, NDBlock):
        old_chunksize = im.get_chunksize()
        fmt = ImageFormat.NDBLOCK
    else:
        raise ValueError(f'Unexpected input type im {type(im)}')

    if storage_options is None:
        preferred_chunksize = old_chunksize
    else:
        preferred_chunksize = storage_options.get('preferred_chunksize') or old_chunksize

    if isinstance(im, np.ndarray):
        NDBlock.save(file, NDBlock(im), storage_options=storage_options)
    elif isinstance(im, da.Array):
        if old_chunksize != preferred_chunksize:
            im = im.rechunk(preferred_chunksize)
        NDBlock.save(file, NDBlock(im), storage_options=storage_options)
    elif isinstance(im, NDBlock):
        if im.get_repr_format() == cvpl_ndblock.ReprFormat.DASK_ARRAY and old_chunksize != preferred_chunksize:
            im = NDBlock(im.get_arr().rechunk(preferred_chunksize))
        NDBlock.save(file, im, storage_options=storage_options)
    else:
        raise ValueError(f'Unexpected input type im {type(im)}')
    with open(f'{file}/.save_meta.txt', mode='w') as outfile:
        outfile.write(str(fmt.value))
        outfile.write(f'\n{chunksize_to_str(old_chunksize)}\n{chunksize_to_str(preferred_chunksize)}')

        compressor = storage_options.get('compressor')
        outfile.write(f'\ncompressor:{repr(compressor)}')


def load(file: str, storage_options: dict = None):
    """Load an image from the given directory.

    The image is one saved by cvpl_tools.im.fs.save()

    Args:
        file: Full path to the directory to be read from
        storage_options: Specifies options in saving method and saved file format

    Returns:
        Recreated image; this method attempts to keep meta and content of the loaded image stays
        the same as when they are saved
    """
    with open(f'{file}/.save_meta.txt') as outfile:
        items = outfile.read().split('\n')
        fmt = ImageFormat(int(items[0]))
        old_chunksize, preferred_chunksize = str_to_chunksize(items[1]), str_to_chunksize(items[2])
    if fmt == ImageFormat.NUMPY:
        im = NDBlock.load(file, storage_options=storage_options).get_arr()
    elif fmt == ImageFormat.DASK_ARRAY:
        im = NDBlock.load(file, storage_options=storage_options).get_arr()
        if old_chunksize != preferred_chunksize:
            im = im.rechunk(old_chunksize)
    elif fmt == ImageFormat.NDBLOCK:
        im = NDBlock.load(file, storage_options=storage_options)
        if im.get_repr_format() == cvpl_ndblock.ReprFormat.DASK_ARRAY and old_chunksize != preferred_chunksize:
            im = NDBlock(im.get_arr().rechunk(old_chunksize))
    else:
        raise ValueError(f'Unexpected input type im {fmt}')
    return im


def display(file: str, viewer_args: dict):
    """Display an image in the viewer; supports numpy or dask ome zarr image

    The image is one saved by cvpl_tools.im.fs.save()

    Args:
        file: Full path to the directory to be read from
        viewer_args: contains viewer and arguments passed to the viewer's add image functions
    """
    viewer_args = copy.copy(viewer_args)
    viewer: napari.Viewer = viewer_args.pop('viewer')
    layer_args = viewer_args.get('layer_args', {})

    with open(f'{file}/.save_meta.txt') as outfile:
        fmt = ImageFormat(int(outfile.read().split('\n')[0]))
    if fmt == ImageFormat.NUMPY:
        is_numpy = True
    elif fmt == ImageFormat.DASK_ARRAY:
        is_numpy = False
    elif fmt == ImageFormat.NDBLOCK:
        properties = NDBlock.load_properties(f'{file}/properties.json')
        repr_fmt: cvpl_ndblock.ReprFormat = properties['repr_format']
        if repr_fmt == cvpl_ndblock.ReprFormat.NUMPY_ARRAY:
            is_numpy = True
        elif repr_fmt == cvpl_ndblock.ReprFormat.NUMPY_ARRAY:
            is_numpy = False
        else:
            raise ValueError(f'Image to be displayed can not be a dict of blocks that is {repr_fmt}')

    is_label: bool = viewer_args.pop('is_label', False)
    if is_numpy:
        fn = viewer.add_labels if is_label else viewer.add_image
        im = NDBlock.load(file).get_arr()
        fn(im, **layer_args)
    else:
        # image saved by NDBlock.save(file)
        add_ome_zarr_array_from_path(viewer, f'{file}/dask_im', use_zip=False, merge_channels=True,
                                     kwargs=layer_args | dict(is_label=is_label))


# -------------------------------------File Caching---------------------------------------------


def cache_im(fn,
             cpath: CachePath,
             cache_level: int | float = 0,
             viewer_args: dict = None):
    """Caches an image object

    Args:
        fn: Computes the image if it's not already cached
        cpath: The cache ID within this directory
        cache_level: cache level of this operation; note even if the caching is skipped, if there is
            a cache file already available on disk then the file will still be read
        viewer_args: contains viewer and arguments passed to the viewer's add image functions

    Returns:
        The cached image loaded
    """
    if viewer_args is None:
        viewer_args = {}
    else:
        viewer_args = copy.copy(viewer_args)  # since we will pop off some attributes

    preferred_chunksize = viewer_args.pop('preferred_chunksize', None)
    multiscale = viewer_args.pop('multiscale', None)
    storage_options = viewer_args.pop('storage_options', {})
    if preferred_chunksize is not None:
        storage_options['preferred_chunksize'] = preferred_chunksize
    if multiscale is not None:
        storage_options['multiscale'] = multiscale

    raw_path = cpath.abs_path
    skip_cache = viewer_args.get('skip_cache', False) or cache_level > viewer_args.get('cache_level', np.inf)
    if not cpath.exists:
        im = fn()
        if skip_cache:
            return im
        save(raw_path, im, storage_options)

    assert os.path.exists(raw_path), f'Directory should be created at path {raw_path}, but it is not found'
    if not skip_cache and viewer_args.get('viewer') is not None:
        viewer_args['layer_args'] = copy.copy(viewer_args.get('layer_args', {}))
        viewer_args['layer_args'].setdefault('name', cpath.cid)
        display(raw_path, viewer_args)

    loaded = load(raw_path, storage_options)

    return loaded


class CachePath:
    """A CachePath class is a pointer to a cached location within a hierarchical directory structure.

    CachePath and CacheDirectory are two classes that implements the file-directory programming pattern,
    where CacheDirectory is a subclass of CachePath and contains zero or more CachePath as its children.
    To create a CachePath object, use the CacheDirectory's cache() function to allocate a new or find
    an existing cache location.

    CachePath as its name suggests is not a file but only a pointer. Creating a CachePath object will not
    create the associated file automatically.
    """

    def __init__(self, root: CacheRootDirectory, path_segs: tuple[str, ...], meta: dict = None,
                 parent: CacheDirectory | None = None, exists: bool = False):
        """Create a CachePath object that manages meta info about the cache file or directory

        Args:
            root: The CacheDirectory that is the root of the directory tree where this file locates under
            path_segs: The path segments associated with this CachePath object, relative to the root
            meta: The meta information associated with this object; will be automatically inferred
                from the path (only able to do so in some situations) if None is provided
            parent: parent of the CachePath object in the cache directory tree; None if this is root
            exists: True if this object references a existing cached object, False if the referenced
                object is newly created
        """
        assert isinstance(path_segs, tuple), f'Expected tuple for path_segs, got {type(path_segs)}'
        self._root = root
        self._path_segs = path_segs

        if meta is None:
            filename = path_segs[-1]
            meta = CachePath.meta_from_filename(filename)

        self._meta = meta
        self._parent = parent
        self._exists = exists
        for key in ('is_dir', 'is_tmp', 'cid'):
            assert key in meta, f'Missing key {key}'

    @property
    def root(self) -> CacheRootDirectory:
        return self._root

    @property
    def parent(self) -> CacheDirectory | None:
        return self._parent

    @property
    def exists(self) -> bool:
        return self._exists

    @property
    def filename(self) -> str:
        """Returns the filename, the last segment of the path

        This is typically cid prepended with is_dir and is_tmp information; do not call this
        on root directory.
        """
        return self._path_segs[-1]

    @property
    def rel_path(self) -> str:
        """Obtain the relative os path to the root"""
        return '/'.join(self._path_segs)

    @property
    def abs_path(self):
        """Obtain the os path under which you can create a directory

        The first time a CachePath object is created for this path, cache_path.abs_path will point to an empty location;
        second time onwards if the directory is not removed, then the returned cache_path.abs_path will point to the
        previously existing directory
        """
        return '/'.join((self._root.abs_path,) + self._path_segs)

    @property
    def is_dir(self):
        """Returns True if this is a directory object instead of a file.

        In other words, this function returns False if this is a leaf node.
        """
        return self._meta['is_dir']

    @property
    def is_tmp(self):
        return self._meta['is_tmp']

    @property
    def cid(self):
        return self._meta['cid']

    @property
    def meta(self):
        return self._meta

    @staticmethod
    def meta_from_filename(file: str, return_none_if_malform=False) -> dict[str, Any] | None:
        """Retrieve meta information from the path

        Args:
            file: filename of the (existing to planning to be created) CachePath object
            return_none_if_malform: If True, return None instead of throwing error if a malformed
                filename is given

        Returns:
            A dictionary of the meta information
        """
        if file.startswith('file_'):
            is_dir = False
            rest = file[len('file_'):]
        elif file.startswith('dir_'):
            is_dir = True
            rest = file[len('dir_'):]
        else:
            if return_none_if_malform:
                return None
            else:
                raise ValueError(f'path is not expected when parsing is_file: {file}')
        if rest.startswith('tmp_'):
            is_tmp = True
            rest = rest[len('tmp_'):]
        elif rest.startswith('cache_'):
            is_tmp = False
            rest = rest[len('cache_'):]
        else:
            if return_none_if_malform:
                return None
            else:
                raise ValueError(f'path is not expected when parsing is_tmp: {file}')
        return dict(
            is_dir=is_dir,
            is_tmp=is_tmp,
            cid=rest
        )

    @staticmethod
    def filename_form_meta(meta: dict[str, Any]) -> str:
        """Obtain filename from the meta dict

        Args:
            meta: The dictionary containing meta information for the CachePath object

        Returns:
            A string as the filename of the cached directory or file
        """
        s1 = 'dir_' if meta['is_dir'] else 'file_'
        s2 = 'tmp_' if meta['is_tmp'] else 'cache_'
        cid = meta['cid']
        return f'{s1}{s2}{cid}'


class CacheDirectory(CachePath):
    """A CacheDirectory is a hierarchical directory structure, corresponding to a directory in the os

    CachePath and CacheDirectory are two classes that implements the file-directory programming pattern.
    """

    def __init__(self,
                 cid: str,
                 root: CacheRootDirectory,
                 path_segs: tuple[str, ...],
                 remove_when_done: bool = True,
                 read_if_exists: bool = True,
                 parent: CacheDirectory | None = None,
                 exists: bool = False):
        """Creates a CacheDirectory instance
        
        Unlike CachePath, calling the __init__ of CacheDirectory will create a physical directory 
        on the disk if one does not already exist.

        Args:
            cid: cid of the directory
            root: The root directory containing this tree of cache directories
            path_segs: The os path to which the directory is to be created; must be empty if read_if_exists=True
            remove_when_done: If True, the entire directory will be removed when it is closed by __exit__; if
                False, then only the temporary folders within the directory will be removed. (The entire subtree
                will be traversed to find any file or directory whose is_tmp is True, and they will be removed)
            read_if_exists: If True, will read from the existing directory at the given path
            parent: parent of the directory in the directory structure; None if is root
            exists: True if the directory is read from an already existing cache; False if it is created anew
        """
        super().__init__(
            root=root,
            path_segs=path_segs,
            meta=dict(
                is_dir=True,
                is_tmp=remove_when_done,
                cid=cid
            ),
            parent=parent,
            exists=exists
        )
        self.cur_idx = 0
        self.read_if_exists = read_if_exists
        self.children: dict[str, CachePath] = {}

        abs_path = self.abs_path
        if self.read_if_exists:
            ensure_dir_exists(abs_path, remove_if_already_exists=False)
            self.children = self.children_from_path(
                prefix_path_segs=self._path_segs)
        else:
            assert not exists, 'when read_if_exists=False, directory must be created so must not already exists, ' \
                               f'please check if any file exists under {abs_path}.'

    def children_from_path(self, prefix_path_segs: tuple[str, ...] = None) -> dict[str, CachePath]:
        """Examine an existing directory path, return recursively all files and directories as dict.

        Args:
            prefix_path_segs: segments prefixing the filenames found under this directory

        Returns:
            Returned json dictionary contains a hierarchical str -> CachePath map; use CachePath.is_dir to
            determine if they contain more children
        """
        children = {}
        abs_path = self.abs_path
        filenames = list(os.listdir(abs_path))
        for filename in filenames:
            subpath_segs = prefix_path_segs + (filename,)
            meta = CachePath.meta_from_filename(filename, return_none_if_malform=True)
            if meta is not None:
                if meta['is_dir']:
                    child = CacheDirectory(
                        cid=meta['cid'],
                        root=self.root,
                        path_segs=subpath_segs,
                        remove_when_done=meta['is_tmp'],
                        read_if_exists=True,
                        parent=self,
                        exists=True
                    )
                    child.children = child.children_from_path(prefix_path_segs=subpath_segs)
                else:
                    child = CachePath(root=self.root, path_segs=subpath_segs, meta=meta, parent=self, exists=True)
                children[meta['cid']] = child
        return children

    def get_children_json(self) -> dict:
        children_json = {}
        for key, child in self.children.items():
            if child.is_dir:
                child: CacheDirectory
                children_json[key] = dict(
                    children=child.get_children_json(),
                    meta=child.meta
                )
            else:
                children_json[key] = child.meta
        return children_json

    def get_children_str(self):
        return json.dumps(self.get_children_json(), indent=2)

    def __getitem__(self, cid: str) -> CachePath | CacheDirectory:
        """Get a CachePath object by its cid"""
        return self.children[cid]

    def __contains__(self, item: str):
        """Checks if an object is cached"""
        return item in self.children

    def _create_cache(self,
                      is_dir=False,
                      cid: str = None
                      ) -> CachePath | CacheDirectory:
        """Return a directory that is guaranteed to be empty within the temporary directory

        This is the interface to create new CachePath or CacheDirectory within this directory.
        The directory will not be immediately created but need to be done manually if is_dir=False.
        When cid=None, the first returned variable (is_cached) will always be False

        Args:
            is_dir: If False, this creates a subfolder that have no children; if True, this creates
                a CacheDirectory recursively
            cid: If specified, will attempt to find cache if already exists; otherwise a temporary
                cache will be returned

        Returns:
            A tuple (is_cached, CachePath), is_cached giving whether the file is cached or is newly
            created. If is_cached is True, then the user should directly read from the cached file
            instead
        """
        is_tmp = cid is None
        if is_tmp:
            cid = f'_{self.cur_idx}'
            self.cur_idx += 1
        else:
            if cid in self.children:
                file = self.children[cid]
                assert file.is_dir == is_dir, f'Unexpected file/directory at {file.abs_path}'
                return self.children[cid]

        # create a new cached object, exists=False
        meta = dict(
            is_dir=is_dir,
            is_tmp=is_tmp,
            cid=cid
        )
        filename = CachePath.filename_form_meta(meta)
        subpath_segs = self._path_segs + (filename,)
        if is_dir:
            cache_path = CacheDirectory(
                cid=cid,
                root=self.root,
                path_segs=subpath_segs,
                remove_when_done=is_tmp,
                read_if_exists=self.read_if_exists,
                parent=self,
                exists=False
            )
        else:
            cache_path = CachePath(
                self.root,
                subpath_segs,
                meta=meta,
                parent=self,
                exists=False
            )
        self.children[cid] = cache_path
        return cache_path
    
    def cache(self, cid: str = None):
        """Similar to _create_cache(), but use an intermediate object to offset the decision of is_dir
        
        Often we pass an empty CacheDirectory object to a function, which saves the intermediate results
        in the empty directory. When the function saves a single file, it will prefer to save to a
        CachePath instead of a CacheDirectory. This makes creating CacheDirectory before function call
        inappropriate. To avoid this, we pass in a struct that contains the parent directory and the
        desired path to create either a CachePath or a CacheDirectory, and offset the decision to
        the function itself.

        Args:
            cid: The cid of either the CachePath or the CacheDirectory

        Returns:
            A CachePointer object of the cache to be created
        """
        return CachePointer(self, cid)

    def cache_subpath(self, cid: str = None) -> CachePath:
        """Wrapper, calls cache(is_dir=False)

        See the docs for CacheDirectory.cache() for more information.
        """
        return self._create_cache(is_dir=False, cid=cid)

    def cache_subdir(self, cid: str = None) -> CacheDirectory:
        """Wrapper, calls cache(is_dir=True)

        See the docs for CacheDirectory.cache() for more information.
        """
        return self._create_cache(is_dir=True, cid=cid)

    def cache_im(self,
                 fn,
                 cid: str = None,
                 cache_level: int | float = 0,
                 viewer_args: dict = None):
        """Caches an image object

        Args:
            fn: Computes the image if it's not already cached
            cid: The cache ID within this directory
            cache_level: cache level of this operation; note even if the caching is skipped, if there is
                a cache file already available on disk then the file will still be read
            viewer_args: contains viewer and arguments passed to the viewer's add image functions

        Returns:
            The cached image loaded
        """
        cpath = self.cache_subpath(cid=cid)
        return cache_im(fn, cpath, cache_level, viewer_args)

    def remove_tmp(self):
        """traverse all subnodes and self, removing those with is_tmp=True"""
        if self.is_tmp:
            shutil.rmtree(self.abs_path)
        else:
            for ch in self.children.values():
                if ch.is_tmp:
                    shutil.rmtree(ch.abs_path)
                elif ch.is_dir:
                    assert isinstance(ch, CacheDirectory)
                    ch.remove_tmp()


class CacheRootDirectory(CacheDirectory):
    def __init__(self,
                 path: str,
                 remove_when_done: bool = True,
                 read_if_exists: bool = True):
        """Creates a CacheDirectory instance

        For users of this class, use this interface to create CacheDirectory

        Args:
            path: The os path to which the directory is to be created; must be empty if read_if_exists=True
            remove_when_done: If True, the entire directory will be removed when it is closed by __exit__; if
                False, then only the temporary folders within the directory will be removed. (The entire subtree
                will be traversed to find any file or directory whose is_tmp is True and remote them)
            read_if_exists: If True, will read from the existing directory at the given path
        """
        self._root_path = path
        super().__init__(
            cid='_RootDirectory',  # not used anywhere else, put a non-empty cid for debugging
            root=self,
            path_segs=tuple(),
            remove_when_done=remove_when_done,
            read_if_exists=read_if_exists,
            parent=None,
            exists=os.path.exists(path)
        )

    @property
    def abs_path(self):
        """Obtain the root directory path itself in the operating system"""
        return self._root_path

    def __enter__(self):
        """Called using the syntax:

        with CacheRootDirectory(...) as cache_dir:
            ...
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.remove_tmp()


class MultiOutputStream:
    def __init__(self, *files):
        self.files = files

    def write(self, message):
        for file in self.files:
            file.write(message)

    def flush(self):
        for file in self.files:
            file.flush()


class CachePointer(CachePath):
    def __init__(self, cdir: CacheDirectory, cid: str | None):
        super().__init__(
            root=cdir.root,
            path_segs=cdir._path_segs + (cid,),
            meta=dict(
                is_tmp=cid is None,
                is_dir=False,
                cid=cid
            ),
            parent=cdir,
            exists=cid is not None and cid in cdir.children
        )

    def subpath(self):
        """Create a CachePath under the parent of this pointer, at the location pointed by this pointer"""
        return self.parent.cache_subpath(cid=self.cid)

    def subdir(self):
        """Create a CacheDirectory under the parent of this pointer, at the location pointed by this pointer"""
        return self.parent.cache_subdir(cid=self.cid)

    def im(self, *args, **kwargs):
        return self.parent.cache_im(cid=self.cid, *args, **kwargs)

    @property
    def is_dir(self):
        return NotImplementedError("A pointer offsets decision of is_dir, don't call is_dir on CachePointer")





"""
Modified from reference: https://github.com/dask/dask-image/blob/main/dask_image/ndinterp/__init__.py

Copyright (c) 2017-2018, dask-image Developers (see AUTHORS.rst for details)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors
may be used to endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import functools
import math
from itertools import product

import dask.array as da
import numpy as np
from dask.base import tokenize
from dask.highlevelgraph import HighLevelGraph
import scipy
from scipy.ndimage import affine_transform as ndimage_affine_transform

from dask_image.dispatch._dispatch_ndinterp import (
    dispatch_affine_transform,
    dispatch_asarray,
    dispatch_spline_filter,
)
from dask_image.ndfilters._utils import _get_depth_boundary


def scale_nearest(image: da.Array, scale: float | tuple[float, ...], output_shape: tuple[int, ...],
                  output_chunks: tuple[int, ...] = None, **kwargs) -> da.Array:
    """Scaling the image without interpolation (order=0)

    Args:
        image: The image to be scaled by the given factor
        scale: The scale to be applied to each axis; if float
        output_shape: Shape of the output array
        output_chunks: Shape of the chunks of the output array
        **kwargs: arguments to be passed to affine_transform_nearest

    Returns:
        Scaled dask array
    """
    if isinstance(scale, int):
        scale = (scale,) * image.ndim + (1,)
    matrix = 1 / np.diag(scale)
    return affine_transform_nearest(image,
                                    matrix,
                                    offset=0.,
                                    output_shape=output_shape,
                                    output_chunks=output_chunks,
                                    **kwargs)


def affine_transform_nearest(
        image: da.Array,
        matrix,
        offset=0.,
        output_shape: tuple[int, ...] = None,
        output_chunks: tuple[int, ...] = None,
        **kwargs
) -> da.Array:
    """Apply an affine transform using Dask. For every
    output chunk, only the slice containing the relevant part
    of the image is processed. Chunkwise processing is performed
    either using `ndimage.affine_transform` or
    `cupyx.scipy.ndimage.affine_transform`, depending on the input type.

    Notes
    -----
        Differences to `ndimage.affine_transformation`:
        - modes 'reflect', 'mirror' and 'wrap' are not supported

        Arguments equal to `ndimage.affine_transformation`,
        except for `output_chunks`.

    Parameters
    ----------
    image : array_like (Numpy Array, Cupy Array, Dask Array...)
        The image array.
    matrix : array (ndim,), (ndim, ndim), (ndim, ndim+1) or (ndim+1, ndim+1)
        Transformation matrix.
    offset : float or sequence, optional
        The offset into the array where the transform is applied. If a float,
        `offset` is the same for each axis. If a sequence, `offset` should
        contain one value for each axis.
    output_shape : tuple of ints, optional
        The shape of the array to be returned.
    output_chunks : tuple of ints, optional
        The shape of the chunks of the output Dask Array.

    Returns
    -------
    affine_transform : Dask Array
        A dask array representing the transformed output

    """
    assert kwargs.get('order', 0) == 0, f'Affine_transform_nearest does not take order parameter!'

    if not isinstance(image, da.core.Array):
        image = da.from_array(image)

    if output_shape is None:
        output_shape = image.shape

    if output_chunks is None:
        output_chunks = image.shape

    # Perform test run to ensure parameter validity.
    ndimage_affine_transform(np.zeros([0] * image.ndim),
                             matrix,
                             offset)

    # Make sure parameters contained in matrix and offset
    # are not overlapping, i.e. that the offset is valid as
    # it needs to be modified for each chunk.
    # Further parameter checks are performed directly by
    # `ndimage.affine_transform`.

    matrix = np.asarray(matrix)
    offset = np.asarray(offset).squeeze()

    # these lines were copied and adapted from `ndimage.affine_transform`
    if (matrix.ndim == 2 and matrix.shape[1] == image.ndim + 1 and
            (matrix.shape[0] in [image.ndim, image.ndim + 1])):
        # assume input is homogeneous coordinate transformation matrix
        offset = matrix[:image.ndim, image.ndim]
        matrix = matrix[:image.ndim, :image.ndim]

    cval = kwargs.pop('cval', 0)
    mode = kwargs.pop('mode', 'constant')
    prefilter = kwargs.pop('prefilter', False)

    supported_modes = ['constant', 'nearest']
    if scipy.__version__ > np.lib.NumpyVersion('1.6.0'):
        supported_modes += ['grid-constant']
    if mode in ['wrap', 'reflect', 'mirror', 'grid-mirror', 'grid-wrap']:
        raise NotImplementedError(
            f"Mode {mode} is not currently supported. It must be one of "
            f"{supported_modes}.")

    n = image.ndim
    image_shape = image.shape

    # calculate output array properties
    normalized_chunks = da.core.normalize_chunks(output_chunks,
                                                 tuple(output_shape))
    block_indices = product(*(range(len(bds)) for bds in normalized_chunks))
    block_offsets = [np.cumsum((0,) + bds[:-1]) for bds in normalized_chunks]

    # use dispatching mechanism to determine backend
    affine_transform_method = dispatch_affine_transform(image)
    asarray_method = dispatch_asarray(image)

    # construct dask graph for output array
    # using unique and deterministic identifier
    output_name = 'affine_transform-' + tokenize(image, matrix, offset,
                                                 output_shape, output_chunks,
                                                 kwargs)
    output_layer = {}
    rel_images = []
    for ib, block_ind in enumerate(block_indices):

        out_chunk_shape = [normalized_chunks[dim][block_ind[dim]]
                           for dim in range(n)]
        out_chunk_offset = [block_offsets[dim][block_ind[dim]]
                            for dim in range(n)]

        out_chunk_edges = np.array([i for i in np.ndindex(tuple([2] * n))]) \
                          * np.array(out_chunk_shape) + np.array(out_chunk_offset)

        # map output chunk edges onto input image coordinates
        # to define the input region relevant for the current chunk
        if matrix.ndim == 1 and len(matrix) == image.ndim:
            rel_image_edges = matrix * out_chunk_edges + offset
        else:
            rel_image_edges = np.dot(matrix, out_chunk_edges.T).T + offset

        rel_image_i = np.min(rel_image_edges, 0)
        rel_image_f = np.max(rel_image_edges, 0)

        # Calculate edge coordinates required for the footprint of the
        # spline kernel according to
        # https://github.com/scipy/scipy/blob/9c0d08d7d11fc33311a96d2ac3ad73c8f6e3df00/scipy/ndimage/src/ni_interpolation.c#L412-L419 # noqa: E501
        # Also see this discussion:
        # https://github.com/dask/dask-image/issues/24#issuecomment-706165593 # noqa: E501
        for dim, s in zip(range(n), image_shape):
            rel_image_i[dim] = np.floor(rel_image_i[dim] + .5)
            rel_image_f[dim] = np.floor(rel_image_f[dim] + .5)

            rel_image_i[dim] = np.clip(rel_image_i[dim], 0, s)
            rel_image_f[dim] = np.clip(rel_image_f[dim], 0, s)

        rel_image_slice = []
        for dim in range(n):
            imin = int(rel_image_i[dim])
            imax = max(imin + 1, int(rel_image_f[dim]))
            rel_image_slice.append(slice(imin, imax))
        rel_image_slice = tuple(rel_image_slice)

        rel_image = image[rel_image_slice]

        print('-----', block_ind, '-----')
        print(rel_image_slice)

        """Block comment for future developers explaining how `offset` is
        transformed into `offset_prime` for each output chunk.
        Modify offset to point into cropped image.
        y = Mx + o
        Coordinate substitution:
        y' = y - y0(min_coord_px)
        x' = x - x0(chunk_offset)
        Then:
        y' = Mx' + o + Mx0 - y0
        M' = M
        o' = o + Mx0 - y0
        """

        offset_prime = offset + np.dot(matrix, out_chunk_offset) - rel_image_i

        output_layer[(output_name,) + block_ind] = (
            affine_transform_method,
            (da.core.concatenate3, rel_image.__dask_keys__()),
            asarray_method(matrix),
            offset_prime,
            tuple(out_chunk_shape),  # output_shape
            None,  # out
            0,
            mode,
            cval,
            False  # prefilter
        )

        rel_images.append(rel_image)

    graph = HighLevelGraph.from_collections(output_name, output_layer,
                                            dependencies=[image] + rel_images)

    meta = dispatch_asarray(image)([0]).astype(image.dtype)

    transformed = da.Array(graph,
                           output_name,
                           shape=tuple(output_shape),
                           # chunks=output_chunks,
                           chunks=normalized_chunks,
                           meta=meta)

    return transformed


# magnitude of the maximum filter pole for each order
# (obtained from scipy/ndimage/src/ni_splines.c)
_maximum_pole = {
    2: 0.171572875253809902396622551580603843,
    3: 0.267949192431122706472553658494127633,
    4: 0.361341225900220177092212841325675255,
    5: 0.430575347099973791851434783493520110,
}


def _get_default_depth(order, tol=1e-8):
    """Determine the approximate depth needed for a given tolerance.

    Here depth is chosen as the smallest integer such that ``|p| ** n < tol``
    where `|p|` is the magnitude of the largest pole in the IIR filter.
    """
    return math.ceil(np.log(tol) / np.log(_maximum_pole[order]))


def spline_filter(
        image,
        order=3,
        output=np.float64,
        mode='mirror',
        output_chunks=None,
        *,
        depth=None,
        **kwargs
):
    if not isinstance(image, da.core.Array):
        image = da.from_array(image)

    # use dispatching mechanism to determine backend
    spline_filter_method = dispatch_spline_filter(image)

    try:
        dtype = np.dtype(output)
    except TypeError:  # pragma: no cover
        raise TypeError(  # pragma: no cover
            "Could not coerce the provided output to a dtype. "
            "Passing array to output is not currently supported."
        )

    if depth is None:
        depth = _get_default_depth(order)

    if mode == 'wrap':
        raise NotImplementedError(
            "mode='wrap' is unsupported. It is recommended to use 'grid-wrap' "
            "instead."
        )

    # Note: depths of 12 and 24 give results matching SciPy to approximately
    #       single and double precision accuracy, respectively.
    boundary = "periodic" if mode == 'grid-wrap' else "none"
    depth, boundary = _get_depth_boundary(image.ndim, depth, boundary)

    # cannot pass a func kwarg named "output" to map_overlap
    spline_filter_method = functools.partial(spline_filter_method,
                                             output=dtype)

    result = image.map_overlap(
        spline_filter_method,
        depth=depth,
        boundary=boundary,
        dtype=dtype,
        meta=image._meta,
        # spline_filter kwargs
        order=order,
        mode=mode,
    )

    return result



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as grid_spec
import scipy.stats


# reference: https://matplotlib.org/matplotblog/posts/create-ridgeplots-in-matplotlib/
def plot_kde(X: list[np.ndarray], xrange: tuple[float, float], ylimit: float):
    colors = ['#0000ff', '#3300cc', '#660099', '#990066', '#cc0033', '#ff0000']
    NPLOTS = len(X)

    gs = (grid_spec.GridSpec(NPLOTS, 1))
    fig = plt.figure(figsize=(8, 6))

    # creating empty list
    ax_objs = []

    for i in range(NPLOTS):
        # creating new axes object and appending to ax_objs
        ax_objs.append(fig.add_subplot(gs[i:i + 1, 0:]))

        xmin, xmax = xrange

        # grabbing x and y data from the kde plot
        x = xmin + np.arange(101) / np.array(100., dtype=np.float32) * (xmax - xmin)
        kde = scipy.stats.gaussian_kde(X[i])
        y = kde.evaluate(x)

        # filling the space beneath the distribution
        ax_objs[-1].fill_between(x, y, color=colors[i])

        # setting uniform x and y lims
        ax_objs[-1].set_xlim(xmin, xmax)
        ax_objs[-1].set_ylim(0, ylimit)

        i += 1

    plt.tight_layout()
    plt.show()



from cvpl_tools.im.fs import CachePointer
from cvpl_tools.im.seg_process import SegProcess, BlockToBlockProcess
import numpy as np
import numpy.typing as npt
import dask.array as da
from cvpl_tools.im.ndblock import NDBlock
import cvpl_tools.im.algorithms as algorithms
import cvpl_tools.im.algs.dask_label as dask_label
from scipy.ndimage import (
    label as instance_label
)


class DirectBSToOS(BlockToBlockProcess):
    IDX_COMPACT = 0
    IDX_UNIQUE = 1

    def __init__(self, is_global: bool = False):
        """Converts a n-dimensional binary mask to an int32 ordinal mask based on connected components.

        If is_global is True, return a global instance mask instead of chunk level instance mask

        Args:
            is_global: If True, label globally instead of at chunk level
        """
        super().__init__(np.int32, is_label=True)
        self.is_global = is_global

    def forward(self,
                im: npt.NDArray | da.Array | NDBlock,
                cptr: CachePointer,
                viewer_args=None
                ) -> npt.NDArray | da.Array | NDBlock:
        if not self.is_global or isinstance(im, np.ndarray):
            return super().forward(im, cptr, viewer_args)

        if viewer_args is None:
            viewer_args = {}

        im = dask_label.label(
            im,
            cptr=cptr,
            output_dtype=np.int32,
            viewer_args=viewer_args | dict(logging=True, client=viewer_args['client'])
        )[0]
        return im

    def np_forward(self, bs: npt.NDArray[np.uint8], block_info=None) -> npt.NDArray[np.int32]:
        lbl_im, nlbl = instance_label(bs)
        return lbl_im


class Watershed3SizesBSToOS(BlockToBlockProcess):
    def __init__(self,
                 size_thres=60.,
                 dist_thres=1.,
                 rst=None,
                 size_thres2=100.,
                 dist_thres2=1.5,
                 rst2=60.):
        super().__init__(np.int32, is_label=True)
        self.size_thres = size_thres
        self.dist_thres = dist_thres
        self.rst = rst
        self.size_thres2 = size_thres2
        self.dist_thres2 = dist_thres2
        self.rst2 = rst2

    def np_forward(self, bs: npt.NDArray[np.uint8], block_info=None) -> npt.NDArray[np.int32]:
        lbl_im = algorithms.round_object_detection_3sizes(bs,
                                                          size_thres=self.size_thres,
                                                          dist_thres=self.dist_thres,
                                                          rst=self.rst,
                                                          size_thres2=self.size_thres2,
                                                          dist_thres2=self.dist_thres2,
                                                          rst2=self.rst2,
                                                          remap_indices=True)
        return lbl_im
    # TODO: better visualization of this stage



"""
Segmentation and post-processing.

This file is for methods generating (dask) single class segmentation masks of binary or ordinal types, the latter of
which is a 0-N segmentation of N objects of the same class in an image.

Methods in this file are designed such that they can be easily swapped out and whose performance is then compared
against each other to manual segmentations over a dataset.

The input to these methods are either input 3d single-channel image of type np.float32, or input image paired with
a deep learning segmentation algorithm. The output may be cell count #, binary mask
(np.uint8) or ordinal mask (np.int32).

Conceptually, we define the following input/output types:
IN - Input Image (np.float32) between min=0 and max=1, this is the brightness dask image as input
BS - Binary Segmentation (3d, np.uint8), this is the binary mask single class segmentation
OS - Ordinal Segmentation (3d, np.int32), this is the 0-N where contour 1-N each denotes an object; also single class
LC - List of Centroids, this contains a list of centroids for each block in the original image
CC - Cell Count Map (3d, np.float64), a cell count number (estimate, can be float) for each block
CD - Cell Density Map (3d, np.float64), this is a down sampled map of the cell density of the brain
ATLAS_MAP - A function that maps from the pixel location within brain to brain regions
ATLAS_CD - Atlas, this summarizes cell density about the brain grouped by region of the brain

Each type above may have associated parameters of them that is not changed.
And we work with the following types of methods as steps to obtain final segmentations from the input mask:
preprocess: IN -> IN
e.g. gaussian_blur

predict_bs: IN -> BS
e.g. cellseg3d_predict, simple_threshold

predict_lc: IN -> LC
e.g. blob_dog

predict_cc: IN -> CC
e.g. scaled_sum_intensity

binary_to_inst: BS -> OS
e.g. direct_bs_to_os, watershed

binary_and_centroids_to_inst: (BS, LC) -> OS
e.g. in_contour_voronoi

os_to_lc: OS -> LC
e.g. direct_os_to_centroids

count_from_lc: LC -> CC
e.g. count_lc_ncentroid, count_edge_penalized_lc_ncentroid

count_from_os: OS -> CC
e.g. count_os_ncontour, count_edge_penalized_os_ncontour, count_os_byvolume

count_to_atlas_cell_density: CC -> ATLAS_CD
e.g. average_by_region

Each method is an object implementing the SegStage interface that has the following methods:
- forward(*args, cid=None, viewer=None) -> out

About viewer visualization: They need not be interpretable for large images but should handle small sized dask
images for debugging purpose.
"""

import abc
import logging
from typing import Callable, Any, Sequence

import cvpl_tools.im.fs as imfs
import cvpl_tools.im.algorithms as algorithms
from cvpl_tools.im.ndblock import NDBlock
import dask.array as da
from dask.distributed import print as dprint
import napari
import numpy as np
import numpy.typing as npt

import skimage
from scipy.ndimage import (
    gaussian_filter as gaussian_filter,
    label as instance_label,
    find_objects as find_objects
)


# ------------------------------------Helper Functions---------------------------------------


def lc_interpretable_napari(layer_name: str,
                            lc: npt.NDArray,
                            viewer: napari.Viewer,
                            ndim: int,
                            extra_features: Sequence,
                            text_color: str = 'green'):
    """This function is used to display feature points for LC-typed output

    Args:
        layer_name: displayed name of the layer
        lc: The list of features, each row of length (ndim + nextra)
        viewer: Napari viewer to add points to
        ndim: dimension of the image
        extra_features: extra features to be displayed as text
        text_color: to be used as display text color
    """
    # reference: https://napari.org/stable/gallery/add_points_with_features.html
    nextra = len(extra_features)
    assert isinstance(lc, np.ndarray), 'lc should be of type np.ndarray!'
    assert lc.ndim == 2, (f'Wrong dimension for list of centroids, expected ndim=2, but got lc={lc} and '
                          f'lc.shape={lc.shape}')
    assert lc.shape[1] == nextra + ndim, (f'Wrong number of features for list of centroids, expected length along '
                                          f'first dimension to be nextra+ndim={nextra + ndim} but got '
                                          f'lc.shape={lc.shape}')

    features = {
        extra_features[i]: lc[:, ndim + i] for i in range(nextra)
    }

    strings = [extra_features[i] + '={' + extra_features[i] + ':.2f}' for i in range(nextra)]
    text_parameters = {
        'string': '\n'.join(strings),
        'size': 9,
        'color': text_color,
        'anchor': 'center',
    }
    import time
    stime = time.time()
    viewer.add_points(lc[:, :ndim],
                      size=1.,
                      ndim=ndim,
                      name=layer_name,
                      features=features,
                      text=text_parameters,
                      visible=False)


# ---------------------------------------Interfaces------------------------------------------


logger = logging.getLogger('SEG_PROCESSES')


# class CacheType(enum.Enum):
#     """Useful for determining the amount and what kinds of cache we need"""
#     TEMPORARY_FILE = 0
#     INTERMEDIATE_FILE = 1
#     VIEWER_FILE = 2  # images to be viewed by napari Viewer or other kinds of visualization
#     INTERMEDIATE_RESULT = 3
#     FINAL_RESULT = 4


class SegProcess(abc.ABC):
    def __init__(self):
        pass

    @abc.abstractmethod
    def forward(self, *args, **kwargs) -> Any:  # removed viewer explicitly since that seems to cause issue in IDE docs
        """Compute outputs from inputs (args)

        Args:
            *args: args inputs to forward()
            viewer_args: If not specified, this is just a forward step; if specified, visualization of the step
                underlying working mechanism will be displayed in Napari viewer object viewer_args['viewer'] for
                interpretation of results and debugging purposes
            **kwargs: kwargs inputs to forward

        Returns:
            computed output
        """
        raise NotImplementedError("SegProcess base class does not implement forward()!")


def block_to_block_forward(
        np_forward: Callable,
        im: npt.NDArray | da.Array | NDBlock,
        cptr: imfs.CachePointer,
        viewer_args=None,
        out_dtype: np.dtype = None, is_label: bool = False,
        compute_chunk_sizes: bool = False):
    if viewer_args is None:
        viewer_args = {}

    def compute():
        nonlocal out_dtype, compute_chunk_sizes
        if isinstance(im, np.ndarray):
            result = np_forward(im)
        elif isinstance(im, da.Array):
            assert im is not None
            if out_dtype is None:
                out_dtype = im.dtype
            result = im.map_blocks(
                np_forward,
                meta=np.array(tuple(), dtype=out_dtype),
                dtype=out_dtype
            )
            if compute_chunk_sizes:
                result.compute_chunk_sizes()
        elif isinstance(im, NDBlock):
            assert im is not None
            if out_dtype is None:
                out_dtype = im.get_dtype()
            result = NDBlock.map_ndblocks([im],
                                          np_forward,
                                          out_dtype=out_dtype,
                                          use_input_index_as_arrloc=0)
            if not result.is_numpy() and compute_chunk_sizes:
                result = result.arr
                result.compute_chunk_sizes()
                result = NDBlock(result)
        else:
            raise TypeError(f'Invalid im type: {type(im)}')
        return result

    patched = viewer_args
    if is_label is not None:
        patched |= dict(is_label=is_label)
    result = cptr.im(fn=compute,
                     cache_level=1,
                     viewer_args=patched)
    return result


class BlockToBlockProcess(SegProcess):
    def __init__(self, out_dtype: np.dtype = None, is_label=False, compute_chunk_sizes=False):
        super().__init__()
        self.out_dtype = out_dtype
        self.is_label = is_label
        self.compute_chunk_sizes = compute_chunk_sizes

    def set_out_dtype(self, out_dtype: np.dtype | None):
        self.out_dtype = out_dtype

    def set_is_label(self, is_label: bool | None):
        self.is_label = is_label

    def forward(self, im: npt.NDArray | da.Array | NDBlock, cptr: imfs.CachePointer, viewer_args=None) \
            -> npt.NDArray | da.Array | NDBlock:
        return block_to_block_forward(
            np_forward=self.np_forward,
            im=im,
            cptr=cptr,
            viewer_args=viewer_args,
            out_dtype=self.out_dtype,
            is_label=self.is_label,
            compute_chunk_sizes=self.compute_chunk_sizes
        )

    @abc.abstractmethod
    def np_forward(self, im: npt.NDArray, block_info=None) -> npt.NDArray:
        raise NotImplementedError('Subclass of BlockToBlockProcess should implement np_forward!')


# ---------------------------------------Preprocess------------------------------------------


class GaussianBlur(BlockToBlockProcess):
    def __init__(self, sigma: float):
        super().__init__(np.float32, is_label=False)
        self.sigma = sigma

    def np_forward(self, im: npt.NDArray[np.float32], block_info=None) -> npt.NDArray[np.float32]:
        return gaussian_filter(im, sigma=self.sigma)


# -------------------------------------Predict Binary----------------------------------------


class BSPredictor(BlockToBlockProcess):
    def __init__(self, pred_fn: Callable):
        super().__init__(np.uint8, is_label=True)
        self.pred_fn = pred_fn

    def np_forward(self, im: npt.NDArray[np.float32], block_info=None) -> npt.NDArray[np.uint8]:
        return self.pred_fn(im)


class SimpleThreshold(BlockToBlockProcess):
    def __init__(self, threshold: float):
        super().__init__(np.uint8, is_label=True)
        self.threshold = threshold

    def np_forward(self, im: npt.NDArray[np.float32], block_info=None) -> npt.NDArray[np.uint8]:
        return (im > self.threshold).astype(np.uint8)


# --------------------------------Predict List of Centroids-----------------------------------


class BlobDog(SegProcess):
    def __init__(self, min_sigma=1, max_sigma=2, threshold: float = 0.1, reduce=False):
        super().__init__()
        self.min_sigma = min_sigma
        self.max_sigma = max_sigma
        self.threshold = threshold
        self.reduce = reduce

    def np_features(self, block: npt.NDArray[np.float32], block_info=None) -> npt.NDArray[np.float32]:
        if block_info is not None:
            slices = block_info[0]['array-location']
            lc = skimage.feature.blob_dog(np.array(block * 255, dtype=np.uint8),
                                          min_sigma=self.min_sigma,
                                          max_sigma=self.max_sigma,
                                          threshold=self.threshold).astype(np.float64)  # N * (ndim + 1) ndarray
            start_pos = np.array([slices[i].start for i in range(len(slices))], dtype=np.float64)
            lc[:, :block.ndim] += start_pos[None, :]
            return lc
        else:
            return block

    def feature_forward(self, im: npt.NDArray[np.float32] | da.Array) -> NDBlock[np.float64]:
        return NDBlock.map_ndblocks([NDBlock(im)], self.np_features, out_dtype=np.float64)

    def forward(self,
                im: npt.NDArray[np.float32] | da.Array,
                cptr: imfs.CachePointer,
                viewer_args: dict = None
                ) -> NDBlock:
        if viewer_args is None:
            viewer_args = {}
        viewer = viewer_args.get('viewer', None)
        ndblock = cptr.im(lambda: self.feature_forward(im), cache_level=1)

        if viewer and viewer_args.get('display_points', True):
            blobdog = ndblock.reduce(force_numpy=True)
            lc_interpretable_napari('blobdog_centroids', blobdog, viewer, im.ndim, ['sigma'])

        ndblock = ndblock.select_columns(slice(im.ndim))
        if self.reduce:
            ndblock = ndblock.reduce(force_numpy=False)
        return ndblock


# -------------------------------Direct Cell Count Prediction----------------------------------


class SumScaledIntensity(SegProcess):
    def __init__(self, scale: float = .008, min_thres: float = 0., reduce: bool = True,
                 spatial_box_width: int | None = None):
        """Summing up the intensity and scale it to obtain number of cells, directly

        Args:
            scale: Scale the sum of intensity by this to get number of cells
            min_thres: Intensity below this threshold is excluded (set to 0 before summing)
            reduce: reduce result when calling forward
            spatial_box_width: If not None, will use this as the box width for adding points to Napari
        """
        assert scale >= 0, f'{scale}'
        assert 0. <= min_thres <= 1., f'{min_thres}'
        super().__init__()
        self.scale = scale
        self.min_thres = min_thres
        self.reduce = reduce
        self.spatial_box_width = spatial_box_width

    def np_features(self, block: npt.NDArray[np.float32], block_info=None, spatial_box_width=None) \
            -> npt.NDArray[np.float64]:
        if block_info is not None:
            slices = block_info[0]['array-location']
            if spatial_box_width is not None:
                padded_block = algorithms.pad_to_multiple(block, spatial_box_width)
            else:
                padded_block = block
            masked = padded_block * (padded_block > self.min_thres)
            if slices is not None:
                startpoint = np.array([slices[i].start for i in range(len(slices))],
                                      dtype=np.float64)
            else:
                startpoint = np.zeros((block.ndim,), dtype=np.float64)

            if spatial_box_width is not None:
                subblock_shape = (spatial_box_width,) * block.ndim
                masked = algorithms.np_map_block(masked, block_sz=subblock_shape)
                masked: npt.NDArray = masked.sum(axis=tuple(range(block.ndim, block.ndim * 2)))

                features = np.zeros((masked.size, block.ndim + 1), dtype=np.float64)
                inds = np.array(np.indices(masked.shape, dtype=np.float64))
                transpose_axes = tuple(range(1, block.ndim + 1)) + (0,)
                inds = inds.transpose(transpose_axes).reshape(-1, block.ndim)
                features[:, -1] = masked.flatten() * self.scale
                features[:, :-1] = startpoint[None, :] + (inds + .5) * np.array(subblock_shape, dtype=np.float64)
            else:
                features = np.zeros((1, block.ndim + 1), dtype=np.float64)
                features[0, -1] = masked.sum() * self.scale
                features[:, :-1] = startpoint[None, :] + np.array(block.shape, dtype=np.float64) / 2
            return features
        else:
            return block

    def feature_forward(self, im: npt.NDArray[np.float32] | da.Array, spatial_block_width: int = None) \
            -> NDBlock[np.float64]:
        def map_fn(b, block_info):
            return self.np_features(b, block_info, spatial_block_width)

        return NDBlock.map_ndblocks([NDBlock(im)], map_fn, out_dtype=np.float64)

    def forward(self, im: npt.NDArray | da.Array, cptr: imfs.CachePointer, viewer_args: dict = None) \
            -> NDBlock | npt.NDArray:
        if viewer_args is None:
            viewer_args = {}
        viewer = viewer_args.get('viewer', None)

        cdir = cptr.subdir()
        forwarded = cdir.cache_im(
            fn=lambda: self.feature_forward(im),
            cid='forwarded',
            cache_level=2
        )

        if viewer:
            mask = cdir.cache_im(
                fn=lambda: im > self.min_thres,
                cid='ssi_mask',
                cache_level=2,
                viewer_args=viewer_args | dict(is_label=True))
            if self.spatial_box_width is not None and viewer_args.get('display_points', True):
                ssi = cdir.cache_im(
                    fn=lambda: self.feature_forward(im, spatial_block_width=self.spatial_box_width).reduce(
                        force_numpy=True),
                    cid='ssi',
                    cache_level=2
                )
                lc_interpretable_napari('ssi_block', ssi, viewer, im.ndim, ['ncells'])

            aggregate_ndblock: NDBlock[np.float64] = cdir.cache_im(
                fn=lambda: map_ncell_vector_to_total(forwarded),
                cid='aggregate_ndblock',
                cache_level=2
            )
            heatmap_cptr = cdir.cache(cid='cell_density_map')
            chunk_size = NDBlock(im).get_chunksize()
            heatmap_logging(aggregate_ndblock, heatmap_cptr, viewer_args, chunk_size)

        def fn():
            ndblock = forwarded.select_columns([-1])
            if self.reduce:
                ndblock = ndblock.reduce(force_numpy=False)
            return ndblock

        ssi_result = cdir.cache_im(fn=fn, cid='ssi_result', cache_level=1)

        return ssi_result


# --------------------Convert Binary and Centroid list to Instance Mask------------------------


class BinaryAndCentroidListToInstance(SegProcess):
    """Defines a SegProcess

    This class' instance forward() takes two inputs: The binary mask segmenting objects, and centroids
    detected by the blobdog algorithm. Then it splits the pixels in the binary mask into instance
    segmentation based on the centroids found by blobdog. The result is a more finely segmented mask
    where objects closer together are more likely to be correctly segmented as two
    """

    def __init__(self, maxSplit: int = 10):
        """Initialize a BinaryAndCentroidListToInstance object

        Args:
            maxSplit: If a contour has number of corresponding centroids above (>) this number in lc,
                then the contour is left as is; this parameter exists for optimization purpose, since
                the larger contours have a time complexity O(N * S) to its spatial size S and the
                number of contours N
        """
        super().__init__()
        self.maxSplit = maxSplit

    def split_ndarray_by_centroid(
            self,
            centroids: list[npt.NDArray[np.int64]],
            indices: list[int],
            X: tuple[npt.NDArray]
    ) -> npt.NDArray[np.int32]:
        N = len(centroids)
        assert N >= 2
        assert N == len(indices)
        arr_shape = X[0].shape
        indices = np.array(indices, dtype=np.int32)
        if N < 10:
            X = np.array(X)

            idxD = np.zeros(arr_shape, dtype=np.int32)
            minD = np.ones(arr_shape, dtype=np.float32) * 1e10
            for i in range(N):
                centroid = centroids[i]
                D = X - np.expand_dims(centroid, list(range(1, X.ndim)))
                D = np.linalg.norm(D.astype(np.float32), axis=0)  # euclidean distance
                new_mask = D < minD
                idxD = new_mask * indices[i] + ~new_mask * idxD
                minD = new_mask * D + ~new_mask * minD
            return idxD
        else:
            centroids = np.array(centroids, dtype=np.int32)
            idxD = algorithms.voronoi_ndarray(arr_shape, centroids)
            return indices[idxD]

    def bacl_forward(self,
                     bs: npt.NDArray[np.uint8],
                     lc: npt.NDArray[np.float64],
                     block_info=None) -> npt.NDArray[np.int32]:
        """For a numpy block and list of centroids in block, return segmentation based on centroids"""

        assert isinstance(bs, np.ndarray) and isinstance(lc, np.ndarray), \
            f'Error: inputs must be numpy for the forward() of this class, got bs={type(bs)} and lc={type(lc)}'

        # first sort each centroid into contour they belong to
        input_slices = block_info[0]['array-location']
        lc = lc.astype(np.int64) - np.array(tuple(s.start for s in input_slices), dtype=np.int64)[None, :]
        lbl_im, max_lbl = instance_label(bs)

        lbl_im: npt.NDArray[np.int32] = lbl_im.astype(np.int32)
        max_lbl: int

        # Below, index 0 is background - centroids fall within this are discarded
        contour_centroids = [[] for _ in range(max_lbl + 1)]

        for centroid in lc:
            c_ord = int(lbl_im[tuple(centroid)])
            contour_centroids[c_ord].append(centroid)

        # now we compute the contours, and brute-force calculate what centroid each pixel is closest to
        object_slices = list(find_objects(lbl_im))
        new_lbl = max_lbl + 1

        for i in range(1, max_lbl + 1):
            slices = object_slices[i - 1]
            if slices is None:
                continue

            # if there are 0 or 1 centroid in the contour, we do nothing
            centroids = contour_centroids[i]  # centroids fall within the current contour
            ncentroid = len(centroids)
            if ncentroid <= 1 or ncentroid > self.maxSplit:
                continue

            # otherwise, divide the contour and map pixels to each
            indices = [i] + [lbl for lbl in range(new_lbl, new_lbl + ncentroid - 1)]
            new_lbl += ncentroid - 1
            mask = lbl_im[slices] == i
            stpt = np.array(tuple(s.start for s in slices), dtype=np.int64)
            centroids = [centroid - stpt for centroid in centroids]
            divided = algorithms.coord_map(mask.shape,
                                           lambda *X: self.split_ndarray_by_centroid(centroids, indices, X))

            lbl_im[slices] = lbl_im[slices] * ~mask + divided * mask

        return lbl_im

    def forward(self,
                bs: npt.NDArray[np.uint8] | da.Array,
                lc: NDBlock[np.float64],
                cptr: imfs.CachePointer,
                viewer_args: dict = None
                ) -> npt.NDArray[np.int32] | da.Array:
        cdir = cptr.subdir()
        if viewer_args is None:
            viewer_args = {}
        viewer = viewer_args.get('viewer', None)

        if viewer:
            def compute_lbl():
                if isinstance(bs, np.ndarray):
                    lbl_im = instance_label(bs)[0]
                else:
                    lbl_im = bs.map_blocks(
                        lambda block: instance_label(block)[0].astype(np.int32), dtype=np.int32
                    )
                return lbl_im

            lbl_im = cdir.cache_im(fn=compute_lbl,
                                   cid='before_split',
                                   cache_level=2,
                                   viewer_args=viewer_args | dict(is_label=True))

        def compute_result():
            nonlocal bs, lc
            bs = NDBlock(bs)
            is_numpy = bs.is_numpy() or lc.is_numpy()
            if is_numpy:
                # if one is numpy but not both numpy, force both of them to be numpy
                if not lc.is_numpy():
                    lc = lc.reduce(force_numpy=True)
                elif not bs.is_numpy():
                    bs = bs.as_numpy()

            ndblock = NDBlock.map_ndblocks([bs, lc], self.bacl_forward,
                                           out_dtype=np.int32, use_input_index_as_arrloc=0)
            if is_numpy:
                result = ndblock.as_numpy()
            else:
                tmp_path = cdir.cache_subpath(cid='blocks').abs_path
                result = ndblock.as_dask_array(tmp_path)
            return result

        result = cdir.cache_im(fn=compute_result,
                               cid='result',
                               cache_level=1,
                               viewer_args=viewer_args | dict(is_label=True))
        return result


def map_ncell_vector_to_total(ndblock: NDBlock[np.float64]) -> NDBlock[np.float64]:
    """Aggregate the counts in ncell vector to get a single ncell estimate

    Args:
        ndblock: Each block contains a ncell vector

    Returns:
        The summed ncell by block; coordinates of each ncell is the center of the block
    """

    def map_fn(block: npt.NDArray[np.float64], block_info):
        slices = block_info[0]['array-location']
        midpoint = np.array(tuple((s.stop - s.start + 1) / 2 + s.start for s in slices), dtype=np.float64)
        ndim = midpoint.shape[0]
        result = np.zeros((1, ndim + 1), dtype=np.float64)
        result[0, :-1] = midpoint
        result[0, -1] = block[:, -1].sum()
        return result

    return NDBlock.map_ndblocks([ndblock], map_fn, out_dtype=np.float64)


def heatmap_logging(aggregate_ndblock: NDBlock[np.float64],
                    cptr: imfs.CachePointer,
                    viewer_args: dict,
                    chunk_size: tuple):
    cdir = cptr.subdir()
    np_arr_path = f'{cdir.abs_path}/density_map.npy'
    if not cdir.exists:
        block = aggregate_ndblock.select_columns([-1])
        ndim = block.get_ndim()

        def map_fn(block: npt.NDArray[np.float64], block_info=None):
            return block.reshape((1,) * ndim)

        block = block.map_ndblocks([block], map_fn, out_dtype=np.float64)
        block = block.as_numpy()
        np.save(np_arr_path, block)

    if viewer_args is None:
        viewer_args = {}
    viewer: napari.Viewer = viewer_args.get('viewer', None)

    block = np.load(np_arr_path)
    if viewer:
        block = np.log2(block + 1.)
        viewer.add_image(block, name='cell_density_map', scale=chunk_size, blending='additive', colormap='red',
                         translate=tuple(sz / 2 for sz in chunk_size))



"""
This file defines some additional utilities function for Python dict class, especially for the
purpose of passing parameter settings hierarchically into sub-functions

For completeness, some functions are Python built-in and are not defined here:
dictionary merge (overwrite, to be precise): dictA | dictB
"""
from typing import Iterable, Any


def subset(d: dict, key_set: Iterable, inplace=False) -> dict:
    """Removes keys in key_set from d

    Args:
        d: The input dictionary
        key_set: Set of keys to remove
        inplace: If True, the operation is performed in place

    Returns:
        The result dictionary; if key_set is ordered, then the result dictionary's keys are ordered in
        that order as well.
    """
    if inplace:
        key_set = set(key_set)
        for key in d:
            if key not in key_set:
                d.pop(key)
        for key in key_set:
            assert key in d, f'Subset key {key} does not exist in dictionary d!'
        return d
    else:
        return {key: d[key] for key in key_set}


def add(d1: dict, d2: dict, inplace=False, allow_overlap=False):
    if not allow_overlap:
        overlap = set(d1.keys()) & set(d2.keys())
        assert overlap == set(), f'Key overlapped: {overlap}'
    if inplace:
        d1 |= d2
        return d1
    else:
        return d1 | d2


def minus(d: dict, key_set, inplace=False, allow_missing=True):
    """Removing a set of keys from dictionary

    Args:
        d: The original dictionary
        key_set: Keys to be removed
        inplace: If True, remove the keys from the original dict; otherwise return a new dict
        allow_missing: If False, will throw an error if key_set has a key not found in d

    Returns:
        The resulting dictionary
    """
    if inplace:
        for key in key_set:
            if key in d:
                d.pop(key)
            else:
                assert allow_missing, f'The key to be subtracted {key} is missing from the dictionary d!'
        return d
    else:
        key_set = set(key_set)
        nminus = len(key_set)
        result = {}
        for key in d:
            if key not in key_set:
                result[key] = d[key]
            else:
                nminus -= 1
        assert allow_missing or nminus == 0, f'Some keys to be subtracted are not found in the dictionary'
        return result


def minus_single(d: dict, key: Any, inplace=False, allow_missing=True):
    """Similar to minus(), but with only one key to be subtracted"""
    if inplace:
        if key in d:
            d.pop(key)
        else:
            assert allow_missing, f'Key {key} not found in dictionary!'
        return d
    else:
        result = {key: value for key, value in d.items()}
        if key in result:
            result.pop(key)
        else:
            assert allow_missing, f'Key {key} not found in dictionary!'
        return result



from typing import Sequence

from cvpl_tools.im import seg_process
from cvpl_tools.im.fs import CachePointer
from cvpl_tools.im.seg_process import SegProcess, lc_interpretable_napari, heatmap_logging, map_ncell_vector_to_total
import numpy as np
import numpy.typing as npt
from cvpl_tools.im.ndblock import NDBlock
import cvpl_tools.ome_zarr.io as cvpl_ome_zarr_io


class CountLCBySize(SegProcess):
    """Counting list of cells by size

    Several features:
    1. A size threshold, below which each contour is counted as a single cell (or part of a single cell,
    in the case it is neighbor to boundary of the image)
    2. Above size threshold, the contour is seen as a cluster of cells an estimate of cell count is given
    based on the volume of the contour
    3. For cells on the boundary location, their estimated ncell is penalized according to the distance
    between the cell centroid and the boundary of the image; if the voxels of the cell do not touch
    edge, this penalty does not apply
    4. A min_size threshold, below (<=) which the contour is simply discarded because it's likely just
    an artifact
    """

    def __init__(self,
                 size_threshold: int | float = 25.,
                 volume_weight: float = 6e-3,
                 border_params: tuple[float, float, float] = (3., -.5, 2.),
                 min_size: int | float = 0,
                 reduce: bool = False):
        super().__init__()
        self.size_threshold = size_threshold
        self.volume_weight = volume_weight
        self.border_params = border_params
        self.min_size = min_size
        self.reduce = reduce

    def cc_list(self, lc: npt.NDArray[np.float64], ndim: int, os_shape: tuple) -> npt.NDArray[np.float64]:
        """Assumption: lc[:, ndim] is nvoxel and lc[:, ndim + 1] is edge_contact"""
        ncells = {}
        dc = []
        dc_idx_to_centroid_idx = {}
        for i in range(lc.shape[0]):
            nvoxel = lc[i, ndim].item()
            if nvoxel <= self.min_size:
                ncells[i] = 0.
            else:
                ncells[i] = 0.

                # if no voxel touch the boundary, we do not want to apply the edge penalty
                on_edge = lc[i, ndim + 1].item()
                if on_edge:
                    dc_idx_to_centroid_idx[len(dc)] = i
                    dc.append(lc[i, :ndim])
                else:
                    ncells[i] = 1

                if nvoxel > self.size_threshold:
                    ncells[i] += (nvoxel - self.size_threshold) * self.volume_weight
        ps = CountLCEdgePenalized(os_shape, self.border_params)

        if len(dc) == 0:
            dc_centroids = np.zeros((0, ndim), dtype=np.float64)
        else:
            dc_centroids = np.array(dc, dtype=np.float64)
        dc_ncells = ps.cc_list(dc_centroids, (0,) * ndim)
        for dc_idx in dc_idx_to_centroid_idx:
            i = dc_idx_to_centroid_idx[dc_idx]
            ncells[i] += dc_ncells[dc_idx]
        ncells = np.array([ncells[i] for i in range(len(ncells))], dtype=np.float64)
        return ncells

    def np_features(self,
                    lc: npt.NDArray[np.float64],
                    block_info=None) -> npt.NDArray[np.float64]:
        if block_info is None:
            return np.zeros(tuple(), dtype=np.float64)

        slices = block_info[0]['array-location']
        ndim = len(slices)
        os_shape = tuple(s.stop for s in slices)
        cc_list = self.cc_list(lc, ndim, os_shape)
        features = np.concatenate((lc[:, :ndim], cc_list[:, None]), axis=1)
        return features

    def feature_forward(self, lc: NDBlock[np.float64]) -> NDBlock[np.float64]:
        return NDBlock.map_ndblocks([NDBlock(lc)], self.np_features, out_dtype=np.float64)

    def forward(self,
                lc: NDBlock[np.float64],
                ndim: int,
                cptr: CachePointer,
                viewer_args: dict = None
                ) -> npt.NDArray[np.float64]:
        if viewer_args is None:
            viewer_args = {}
        viewer = viewer_args.get('viewer', None)
        cdir = cptr.subdir()
        ndblock = cdir.cache_im(fn=lambda: self.feature_forward(lc),
                                cid='lc_by_size_features',
                                cache_level=1)

        dp = viewer_args.get('display_points', True)
        if viewer:
            if dp:
                features = ndblock.reduce(force_numpy=True)
                features = features[features[:, -1] > 0., :]
                lc_interpretable_napari('bysize_ncells',
                                        features, viewer, ndim, ['ncells'])

            aggregate_ndblock: NDBlock[np.float64] = cdir.cache_im(
                fn=lambda: map_ncell_vector_to_total(ndblock), cid='aggregate_ndblock',
                cache_level=2
            )
            if dp:
                aggregate_features: npt.NDArray[np.float64] = cdir.cache_im(
                    fn=lambda: aggregate_ndblock.reduce(force_numpy=True), cid='block_cell_count',
                    cache_level=2
                )
                lc_interpretable_napari('block_cell_count', aggregate_features, viewer,
                                        ndim, ['ncells'], text_color='red')

            chunk_size = lc.get_chunksize()
            heatmap_logging(aggregate_ndblock, cdir.cache(cid='cell_density_map'), viewer_args, chunk_size)

        ndblock = ndblock.select_columns([-1])

        if self.reduce:
            ndblock = ndblock.reduce(force_numpy=False)
        ndblock = ndblock.sum(keepdims=True)
        return ndblock


class CountLCEdgePenalized(SegProcess):
    """From a list of cell centroid locations, calculate a cell count estimate

    You need to provide an image_shape parameter due to the fact that lc does not contain
    information about input image shape

    Each centroid is simply treated as 1 cell when they are sufficiently far from the edge,
    but as they get closer to the edge the divisor becomes >1. and their estimate decreases
    towards 0, since cells near the edge may be double-counted (or triple or more counted
    if at a corner etc.)
    """

    def __init__(self,
                 chunks: Sequence[Sequence[int]] | Sequence[int],
                 border_params: tuple[float, float, float] = (3., -.5, 2.),
                 reduce: bool = False):
        """Initialize a CountLCEdgePenalized object

        Args:
            chunks: Shape of the blocks over each axis
            border_params: Specify how the cells on the border gets discounted. Formula is:
                intercept, dist_coeff, div_max = self.border_params
                mults = 1 / np.clip(intercept - border_dists * dist_coeff, 1., div_max)
                cc_list = np.prod(mults, axis=1)
            reduce: If True, reduce the results into a Numpy 2d array calling forward()
        """
        super().__init__()
        if isinstance(chunks[0], int):
            # Turn Sequence[int] to Sequence[Sequence[int]]
            # assume single numpy block, at index (0, 0, 0)
            chunks = tuple((chunks[i],) for i in range(len(chunks)))
        self.chunks = chunks
        self.numblocks = tuple(len(c) for c in chunks)
        self.border_params = border_params
        self.reduce = reduce

        intercept, dist_coeff, div_max = border_params
        assert intercept >= 1., f'intercept has to be >= 1. as divisor must be >= 1! (intercept={intercept})'
        assert dist_coeff <= 0., (f'The dist_coeff needs to be non-positive so divisor decreases as cell is further '
                                  f'from the edge')
        assert div_max >= 1., f'The divisor is >= 1, but got div_max < 1! (div_max={div_max})'

    def cc_list(self,
                lc: npt.NDArray[np.float64],
                block_index: tuple) -> npt.NDArray[np.float64]:
        """Returns a cell count estimate for each contour in the list of centroids

        Args:
            lc: The list of centroids to be given cell estimates for
            block_index: The index of the block which this lc corresponds to

        Returns:
            A 1-d list, each element is a scalar cell count for the corresponding contour centroid in lc
        """
        block_shape = np.array(
            tuple(self.chunks[i][block_index[i]] for i in range(len(self.chunks))),
            dtype=np.float64
        )
        midpoint = (block_shape * .5)[None, :]

        # compute border distances in each axis direction
        border_dists = np.abs((lc[:, :len(self.chunks)] + midpoint) % block_shape - (midpoint - .5))

        intercept, dist_coeff, div_max = self.border_params
        mults = 1 / np.clip(intercept + border_dists * dist_coeff, 1., div_max)
        cc_list = np.prod(mults, axis=1)
        return cc_list

    def np_features(self, lc: npt.NDArray[np.float64], block_info=None) -> npt.NDArray[np.float64]:
        """Calculate cell counts, then concat centroid locations to the left of cell counts"""
        cc_list = self.cc_list(lc, block_info[0]['chunk-location'])
        features = np.concatenate((lc[:, :len(self.chunks)], cc_list[:, None]), axis=1)
        return features

    def feature_forward(self, lc: NDBlock[np.float64]) -> NDBlock[np.float64]:
        return NDBlock.map_ndblocks([lc], self.np_features, out_dtype=np.float64)

    def forward(self,
                lc: NDBlock[np.float64],
                cptr: CachePointer,
                viewer_args: dict = None) -> NDBlock[np.float64]:
        if viewer_args is None:
            viewer_args = {}
        viewer = viewer_args.get('viewer', None)
        assert lc.get_numblocks() == self.numblocks, ('numblocks could not match up for the chunks argument '
                                                      f'provided, expected {self.numblocks} but got '
                                                      f'{lc.get_numblocks()}')
        cdir = cptr.subdir()
        ndblock = cdir.cache_im(fn=lambda: self.feature_forward(lc), cid='lc_cc_edge_penalized',
                                cache_level=1)

        dp = viewer_args.get('display_points', True)
        if viewer:
            if viewer_args.get('display_checkerboard', True):
                checkerboard = cdir.cache_im(fn=lambda: cvpl_ome_zarr_io.dask_checkerboard(self.chunks),
                                             cid='checkerboard',
                                             cache_level=2,
                                             viewer_args=viewer_args | dict(is_label=True))

            if dp:
                features = ndblock.reduce(force_numpy=True)
                lc_interpretable_napari('lc_cc_edge_penalized', features, viewer,
                                        len(self.chunks), ['ncells'])

            aggregate_ndblock: NDBlock[np.float64] = cdir.cache_im(
                fn=lambda: map_ncell_vector_to_total(ndblock), cid='aggregate_ndblock',
                cache_level=2
            )
            if dp:
                aggregate_features: npt.NDArray[np.float64] = cdir.cache_im(
                    fn=lambda: aggregate_ndblock.reduce(force_numpy=True), cid='block_cell_count',
                    cache_level=2
                )
                lc_interpretable_napari('block_cell_count', aggregate_features, viewer,
                                        len(self.chunks), ['ncells'], text_color='red')

            chunk_size = tuple(ax[0] for ax in self.chunks)
            heatmap_logging(aggregate_ndblock, cdir.cache(cid='cell_density_map'), viewer_args, chunk_size)

        ndblock = ndblock.select_columns([-1])
        if self.reduce:
            ndblock = ndblock.reduce(force_numpy=False)
        ndblock = ndblock.sum(keepdims=True)
        return ndblock



"""
Contains default, quick setup of the pipeline essential objects to run subclasses of SegProcess class
"""
import dataclasses
import os.path

import cvpl_tools.im.fs as imfs
from dask.distributed import Client
import napari


@dataclasses.dataclass
class PLComponents:
    tmp_path: str
    cache_root: imfs.CacheRootDirectory
    dask_client: Client
    viewer: napari.Viewer

    def __init__(self, tmp_path, cachedir_name: str, client_args: dict, viewer_args: dict):
        """Create a PLComponents object

        on __enter__, the instance will set up necessary components for running most SegProcess classes

        Args:
            tmp_path: temporary path where cache directory and dask temporary files will be written
            cachedir_name: name of the cache directory to be created under tmp_path
            client_args: arguments to create dask client
            viewer_args: arguments to create napari viewer
        """

        self._cachedir_name = cachedir_name
        self._dask_config = None
        assert isinstance(client_args, dict), f'Expected dictionary, got {type(client_args)}'
        self._client_args = client_args
        assert isinstance(viewer_args, dict), f'Expected dictionary, got {type(viewer_args)}'
        self._viewer_args = viewer_args

        self.tmp_path = tmp_path
        self.cache_root = None
        self.dask_client = None
        self.viewer = None

    def __enter__(self):
        """Called using the syntax:

        with PLComponents(...) as plcs:
            ...
        """
        import sys
        import cvpl_tools.im.fs as imfs
        import dask
        from dask.distributed import Client
        import napari

        # set standard output and error output to use log file, since terminal output has some issue
        # with distributed print
        logfile_stdout = open('log_stdout.txt', mode='w')
        logfile_stderr = open('log_stderr.txt', mode='w')
        sys.stdout = imfs.MultiOutputStream(sys.stdout, logfile_stdout)
        sys.stderr = imfs.MultiOutputStream(sys.stderr, logfile_stderr)

        imfs.ensure_dir_exists(self.tmp_path, remove_if_already_exists=False)

        self._dask_config = dask.config.set({'temporary_directory': self.tmp_path})
        self._dask_config.__enter__()  # emulate the with clause which is what dask.config.set is used in

        self.cache_root = imfs.CacheRootDirectory(
            f'{self.tmp_path}/{self._cachedir_name}',
            remove_when_done=False,
            read_if_exists=True,
        )
        self.cache_root.__enter__()

        self.dask_client = Client()

        vargs = self._viewer_args
        if vargs.get('use_viewer', True):
            self.viewer = napari.Viewer(ndisplay=2)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.dask_client.close()

        self._dask_config.__exit__(exc_type, exc_val, exc_tb)
        self.cache_root.__exit__(exc_type, exc_val, exc_tb)



from cvpl_tools.im.seg_process import SegProcess, lc_interpretable_napari, heatmap_logging, map_ncell_vector_to_total
import numpy as np
import numpy.typing as npt
from cvpl_tools.im.ndblock import NDBlock
import cvpl_tools.im.process.os_to_lc as os_to_lc
import cvpl_tools.im.process.lc_to_cc as lc_to_cc
import dask.array as da
from cvpl_tools.im.fs import CachePointer


class CountOSBySize(SegProcess):
    """Counting ordinal segmentation contours

    Several features:
    1. A size threshold, below which each contour is counted as a single cell (or part of a single cell,
    in the case it is neighbor to boundary of the image)
    2. Above size threshold, the contour is seen as a cluster of cells an estimate of cell count is given
    based on the volume of the contour
    3. For cells on the boundary location, their estimated ncell is penalized according to the distance
    between the cell centroid and the boundary of the image; if the voxels of the cell do not touch
    edge, this penalty does not apply
    4. A min_size threshold, below (<=) which the contour is simply discarded because it's likely just
    an artifact
    """

    def __init__(self,
                 size_threshold: int | float = 25.,
                 volume_weight: float = 6e-3,
                 border_params: tuple[float, float, float] = (3., -.5, 2.),
                 min_size: int | float = 0,
                 reduce: bool = False,
                 is_global: bool = False):
        super().__init__()
        self.size_threshold = size_threshold
        self.volume_weight = volume_weight
        self.border_params = border_params
        self.min_size = min_size
        self.reduce = reduce

        self.os_to_lc = os_to_lc.DirectOSToLC(min_size=min_size, reduce=False, is_global=is_global,
                                              ex_statistics=['nvoxel', 'edge_contact'])
        self.lc_to_cc = lc_to_cc.CountLCBySize(size_threshold=size_threshold,
                                               volume_weight=volume_weight,
                                               border_params=border_params,
                                               reduce=reduce)

    def forward(self,
                im: npt.NDArray[np.int32] | da.Array,
                cptr: CachePointer,
                viewer_args: dict = None
                ) -> npt.NDArray[np.float64]:
        cdir = cptr.subdir()
        lc = self.os_to_lc.forward(im, cdir.cache(cid='os_to_lc'), viewer_args)
        cc = self.lc_to_cc.forward(lc,  im.ndim, cdir.cache(cid='lc_to_cc'), viewer_args)

        return cc



"""
This file defines SegProcess subclasses which converts any image to any other image of the same type (same type in
terms of the classification of different data types defined in the top of seg_process.py)
"""
import dask.array as da
import numpy.typing as npt
from skimage.transform import resize

from cvpl_tools.im.fs import CachePointer
from cvpl_tools.im.seg_process import (
    SegProcess,
    BlockToBlockProcess,
    block_to_block_forward
)


class DownsamplingByIntFactor(SegProcess):
    """*DEPRECATED*

    use cvpl_tools.im.algs.dask_ndinterp.scale_nearest instead
    """
    def __init__(self, factor: int | tuple = 2):
        """Creates a DownsamplingByIntFactor object

        Args:
            factor: Downsampling factor
        """
        super().__init__()
        self._factor = factor

    def forward(self,
                im: npt.NDArray | da.Array,
                cptr: CachePointer,
                viewer_args: dict = None
                ) -> npt.NDArray | da.Array:
        """Downsample the image and

        The function will display the downsampled image if viewer is provided in viewer_args, so make sure
        you set is_label to True if the image is a label instead of brightness/RGB

        If you want displayed image to be of the same scale of original, set the scale in viewer_args
        correspondingly

        Args:
            im: Image to be downsampled (np or dask array)
            cid: Caching id under the current tmp directory
            viewer_args: Arguments for the viewer

        Returns:
            The down-sampled image
        """
        if viewer_args is None:
            viewer_args = {}
        if isinstance(self._factor, int):
            factor = (self._factor,) * im.ndim
        else:
            factor = self._factor
        slices = tuple(slice(None, None, factor[i]) for i in range(im.ndim))
        im = cptr.im(fn=lambda: im[slices],
                     cache_level=1,
                     viewer_args=viewer_args)
        im.compute_chunk_sizes()
        return im


class UpsamplingByIntFactor(BlockToBlockProcess):
    """*DEPRECATED*

    use cvpl_tools.im.algs.dask_ndinterp.scale_nearest instead
    """

    def __init__(self, factor: int | tuple = 2, order: int = 0):
        """Creates a UpsamplingByIntFactor object

        Args:
            factor: Upsampling factor
            order: interpolation order (see skimage.transform.resize())
        """
        super().__init__(compute_chunk_sizes=True)
        self._factor = factor
        self._order = order

    def np_forward(self, im: npt.NDArray, block_info=None) -> npt.NDArray:
        """Calculate cell counts, then concat centroid locations to the left of cell counts"""
        start_dtype = im.dtype
        if isinstance(self._factor, int):
            factor = (self._factor,) * im.ndim
        else:
            factor = self._factor
        new_shape = tuple(im.shape[i] * factor[i] for i in range(im.ndim))
        im = resize(im, new_shape, order=self._order)
        end_dtype = im.dtype
        assert start_dtype == end_dtype, f'start:{start_dtype} end:{end_dtype} are not the same!'
        return im

    def forward(self,
                im: npt.NDArray | da.Array,
                cptr: CachePointer,
                viewer_args: dict = None
                ) -> npt.NDArray | da.Array:
        """Upsample the image

        The function will display the upsampled image if viewer is provided in viewer_args, so make sure
        you set is_label to True if the image is a label instead of brightness/RGB

        If you want displayed image to be of the same scale of original, set the scale in viewer_args
        correspondingly

        Args:
            im: Image to be upsampled (np or dask array)
            cid: Caching id under the current tmp directory
            viewer_args: Arguments for the viewer

        Returns:
            The up-sampled image
        """
        return block_to_block_forward(
            np_forward=self.np_forward,
            cptr=cptr,
            im=im,
            viewer_args=viewer_args,
            out_dtype=None,
            is_label=viewer_args.get('is_label', False),
            compute_chunk_sizes=True
        )



"""
dask-image's label function encounters memory error when in large dataset. This file defines a distributed, on-disk
version of the label() function of scipy.ndimage
"""
from math import prod

import dask.array as da
import numcodecs
import numpy as np
import numpy.typing as npt

from cvpl_tools.im.ndblock import NDBlock, dumps_numpy, loads_numpy
from cvpl_tools.im.partd_server import SQLiteKVStore, SQLitePartd, SqliteServer
from cvpl_tools.im.fs import CacheDirectory, CachePointer
import os
from scipy.ndimage import label as scipy_label
import partd
import pickle
import cvpl_tools.im.algorithms as cvpl_algorithms
from dask.distributed import print as dprint
from collections import defaultdict


class PairKVStore(SQLiteKVStore):
    def init_db(self):
        if not self.is_exists:
            self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS kv_store (
                id TEXT PRIMARY KEY,
                value1 TEXT,
                value2 TEXT
            )
            ''')
            self.write_row_stmt = """
            INSERT INTO kv_store (id, value1, value2) VALUES (?, ?, NULL)
            ON CONFLICT(id) DO UPDATE SET value2=excluded.value1;
            """

    def read_all(self):
        self.cursor.execute("""
        SELECT id FROM kv_store
        """)
        ids = self.cursor.fetchall()
        for row in ids:
            self.cursor.execute("""
            SELECT value1, value2 FROM kv_store WHERE id = ?
            """, row)
            yield self.cursor.fetchone()


def find_connected_components(edges: set[tuple[int, int]]) -> list[set[int, ...], ...]:
    graph = defaultdict(set)

    for u, v in edges:
        graph[u].add(v)
        graph[v].add(u)

    visited = set()
    components = []

    def dfs(node, component):
        visited.add(node)
        component.add(node)
        for neighbor in graph[node]:
            if neighbor not in visited:
                dfs(neighbor, component)

    for node in graph:
        if node not in visited:
            component = set()
            dfs(node, component)
            components.append(component)

    return components


def label(im: npt.NDArray | da.Array | NDBlock,
          cptr: CachePointer,
          output_dtype: np.dtype = None,
          viewer_args: dict = None
          ) -> npt.NDArray | da.Array | NDBlock:
    """Return (lbl_im, nlbl) where lbl_im is a globally labeled image of the same type/chunk size as the input"""

    cdir = cptr.subdir()

    ndim = im.ndim
    if viewer_args is None:
        viewer_args = {}
    is_logging = viewer_args.get('logging', False)
    compressor = numcodecs.Blosc(cname='lz4', clevel=9, shuffle=numcodecs.Blosc.BITSHUFFLE)
    vargs = dict(compressor=compressor)  # this is for compressing labels of uint8 or int32 types

    if isinstance(im, np.ndarray):
        return scipy_label(im, output=output_dtype)
    is_dask = isinstance(im, da.Array)
    if not is_dask:
        assert isinstance(im, NDBlock)
        im = im.as_dask_array(tmp_dirpath=cdir.abs_path)

    def map_block(block: npt.NDArray, block_info: dict):
        lbl_im = scipy_label(block, output=output_dtype)[0]
        return lbl_im

    def to_max(block: npt.NDArray, block_info: dict):
        return block.max(keepdims=True)

    # compute locally labelled chunks and save their bordering slices
    if is_logging:
        print('Locally label the image')
    locally_labeled = cdir.cache_im(
        lambda: im.map_blocks(map_block, meta=np.zeros(tuple(), dtype=output_dtype)),
        cid='locally_labeled_without_cumsum',
        viewer_args=vargs
    )

    def compute_nlbl_np_arr():
        if is_logging:
            print('Taking the max of each chunk to obtain number of labels')
        locally_labeled_ndblock = NDBlock(locally_labeled)
        new_slices = list(tuple(slice(0, 1) for _ in range(ndim))
                          for _ in NDBlock(locally_labeled_ndblock).get_slices_list())
        nlbl_ndblock_arr = NDBlock.map_ndblocks([locally_labeled_ndblock], fn=to_max, out_dtype=output_dtype,
                                                new_slices=new_slices)
        if is_logging:
            print('Convert number of labels of chunks to numpy array')
        nlbl_np_arr = nlbl_ndblock_arr.as_numpy()
        return nlbl_np_arr

    nlbl_np_arr = cdir.cache_im(fn=compute_nlbl_np_arr, cid='nlbl_np_arr')

    def compute_cumsum_np_arr():
        if is_logging:
            print('Compute prefix sum and reshape back')
        cumsum_np_arr = np.cumsum(nlbl_np_arr)
        return cumsum_np_arr

    cumsum_np_arr = cdir.cache_im(fn=compute_cumsum_np_arr, cid='cumsum_np_arr')
    assert cumsum_np_arr.ndim == 1
    total_nlbl = cumsum_np_arr[-1].item()
    cumsum_np_arr[1:] = cumsum_np_arr[:-1]
    cumsum_np_arr[0] = 0
    cumsum_np_arr = cumsum_np_arr.reshape(nlbl_np_arr.shape)
    if is_logging:
        print(f'total_nlbl={total_nlbl}, Convert prefix sum to a dask array then to NDBlock')
    cumsum_da_arr = da.from_array(cumsum_np_arr, chunks=(1,) * cumsum_np_arr.ndim)

    # Prepare cache file to be used
    if is_logging:
        print('Setting up cache sqlite database')
    cache_file = cdir.cache_subpath(cid='border_slices')
    slices_abs_path = cache_file.abs_path
    os.makedirs(slices_abs_path, exist_ok=True)
    db_path = f'{slices_abs_path}/border_slices.db'

    def create_kv_store():
        kv_store = PairKVStore(db_path)
        return kv_store

    def get_sqlite_partd():
        partd = SQLitePartd(slices_abs_path, create_kv_store=create_kv_store)
        return partd

    if is_logging:
        print('Setting up partd server')
    nappend = im.ndim * 2 * prod(locally_labeled.numblocks)
    storage_options = viewer_args.get('storage_options', dict())

    server = None
    def compute_edge_slices():
        nonlocal server
        server = SqliteServer(slices_abs_path, nappend=nappend, get_sqlite_partd=get_sqlite_partd,
                              port_protocol=storage_options.get('port_protocol', 'tcp'))
        server_address = server.address

        # compute edge slices
        if is_logging:
            print('Computing edge slices, writing to database')

        def compute_slices(block: npt.NDArray, block2: npt.NDArray, block_info: dict = None):
            # block is the local label, block2 is the single element prefix summed number of labels

            client = partd.Client(server_address)
            block_index = list(block_info[0]['chunk-location'])
            block = block + (block != 0).astype(block.dtype) * block2
            for ax in range(block.ndim):
                for face in range(2):
                    block_index[ax] += face
                    indstr = '_'.join(str(index) for index in block_index) + f'_{ax}'
                    sli_idx = face * (block.shape[ax] - 1)
                    sli = np.take(block, indices=sli_idx, axis=ax)
                    client.append({
                        indstr: dumps_numpy(sli, compressor)
                    })
                    block_index[ax] -= face
            client.close()
            return block

        nonlocal locally_labeled
        locally_labeled = da.map_blocks(compute_slices, locally_labeled, cumsum_da_arr,
                                        meta=np.zeros(tuple(), dtype=output_dtype))
        return locally_labeled

    locally_labeled = cdir.cache_im(
        compute_edge_slices,
        cid='locally_labeled_with_cumsum',
        viewer_args=vargs
    )
    if server is not None:
        server.wait_join()

    if is_logging:
        print('Process locally to obtain a lower triangular adjacency matrix')
    read_kv_store = PairKVStore(db_path)
    lower_adj = set()
    for value1, value2 in read_kv_store.read_all():
        if value1 is None or value2 is None:
            continue
        sli1, sli2 = loads_numpy(value1, compressor).flatten(), loads_numpy(value2, compressor).flatten()
        sli = np.stack((sli1, sli2), axis=1)
        tups = cvpl_algorithms.np_unique(sli, axis=0)
        for row in tups.tolist():
            i1, i2 = row
            if i2 < i1:
                tmp = i2
                i2 = i1
                i1 = tmp
            if i1 == 0:
                continue
            assert i1 < i2, f'i1={i1} and i2={i2}!'  # can not be equal because indices are globally unique here
            tup = (i2, i1)
            if tup not in lower_adj:
                lower_adj.add(tup)
    connected_components = find_connected_components(lower_adj)
    if is_logging:
        print('Compute final indices remap array')
    ind_map_np = np.arange(total_nlbl + 1, dtype=output_dtype)
    assigned_mask = np.zeros((total_nlbl + 1), dtype=np.uint8)
    assigned_mask[0] = 1  # we don't touch background class
    comp_i = 0
    while comp_i < len(connected_components):
        comp = connected_components[comp_i]
        comp_i += 1
        for j in comp:
            ind_map_np[j] = comp_i
            assigned_mask[j] = 1
    for i in range(assigned_mask.shape[0]):
        if assigned_mask[i] == 0:
            comp_i += 1
            ind_map_np[i] = comp_i

    read_kv_store.close()

    if is_logging:
        print(f'comp_i={comp_i}, Remapping the indices array to be globally consistent')
    client = viewer_args['client']
    ind_map_scatter = client.scatter(ind_map_np, broadcast=True)

    def local_to_global(block, block_info, ind_map_scatter):
        return ind_map_scatter[block]

    globally_labeled = cdir.cache_im(
        lambda: locally_labeled.map_blocks(func=local_to_global, meta=np.zeros(tuple(), dtype=output_dtype),
                                           ind_map_scatter=ind_map_scatter),
        cid='globally_labeled',
        viewer_args=vargs
    )
    result_arr = globally_labeled
    if not is_dask:
        if is_logging:
            print('converting the result to NDBlock')
        result_arr = NDBlock(result_arr)
    if is_logging:
        print('Function ends')

    im = cdir.cache_im(lambda: result_arr,
                       cid='global_os',
                       cache_level=1,
                       viewer_args=viewer_args | dict(is_label=True))

    return result_arr, comp_i



"""
This file is for cv algorithms
"""
import enum
from typing import Callable, Type
import numpy as np
import numpy.typing as npt
import skimage
from scipy.ndimage import (
    distance_transform_edt as distance_transform_edt,
    label as instance_label,
    find_objects as find_objects
)


def coord_map(im_shape: tuple, map_fn: Callable) -> npt.NDArray:
    """
    Take a function mapping coordinates to pixel values and generate the specified image; np.indices
    is used in the underlying implementation.

    Args:
        im_shape: The shape of the image
        map_fn: The function that maps (numpy arrays of) indices to pixel values

    Returns:
        The image whose pixel values are specified by the function

    Example:
        >>> coord_map((2, 3), lambda Y, X: X - Y)
        array([[ 0,  1,  2],
               [-1,  0,  1]])
    """
    coords = np.indices(im_shape)
    im = map_fn(*coords)
    return im


def np_map_block(im: npt.NDArray, block_sz) -> npt.NDArray:
    """map_block(), but for numpy arrays

    Makes image from shape (Z, Y, X...) into (Zb, Yb, Xb..., Zv, Yv, Xv...) where b are block indices within space and
    v are voxel spatial indices within the block.
    Image size should be divisible by block size.

    Args:
        im: The numpy array of n dimensions to be mapped
        block_sz: the shape of each block

    Returns:
        Expanded array with 2n dimensions in total, first n are block indices and last n are voxel indices
    """
    assert im.ndim == len(block_sz), (f'Got block shape {block_sz} of ndim={len(block_sz)} different from im.ndim='
                                      f'{im.ndim}!')
    expanded_shape = []
    for i in range(im.ndim):
        block_axlen = im.shape[i] // block_sz[i]
        voxel_axlen = block_sz[i]
        assert voxel_axlen * block_axlen == im.shape[i], (f'Got indivisible image shape {im.shape[i]} by block size '
                                                          f'{block_sz[i]} on axis {i}')
        expanded_shape.extend((block_axlen, voxel_axlen))

    ax_order = [2 * i for i in range(im.ndim)] + [1 + 2 * i for i in range(im.ndim)]
    expanded_im = im.reshape(expanded_shape)
    return expanded_im.transpose(ax_order)


def np_unique(lbl_im: npt.NDArray[np.int32], return_index: bool = False,
              return_inverse: bool = False, return_counts: bool = False,
              axis=None, *,
              equal_nan: bool = False) -> npt.NDArray:
    """Before 2.0.0, numpy will return an array of 1d regardless of input shape
    
    This function fixes this by forcing the function signature working the same as version 2
    """
    # this function fixes this by forcing the function signature working the same as version 2
    ret = np.unique(lbl_im, return_index, return_inverse, return_counts, axis, equal_nan=equal_nan)
    if return_inverse and np.__version__ < '2.0.0':
        im_shape = lbl_im.shape
        ret = ret[:-1] + (ret[-1].reshape(im_shape),)
    return ret


def pad_to_multiple(arr: npt.NDArray, n: int) -> npt.NDArray:
    """Numpy, pad an array on each axis to a multiple of n.

    This function ensures no operation is done and original array is returned if the shape is already
    a multiple for each dimension. Other-wise a new array will be created with minimum shape matching
    the requirement and it will be returned.

    Args:
        arr: The array to be padded
        n: Each axis should be padded to a multiple of this number

    Returns:
        The padded array
    """
    pad_width = tuple((0, (n - dim % n) % n) for dim in arr.shape)
    for tup in pad_width:
        if tup != (0, 0):
            break
        return arr
    return np.pad(array=arr,
                  pad_width=pad_width,
                  mode='constant')


def setdiff2d(A, B):
    """An extension of np.setdiff1d to 2d array

    copied from
    https://stackoverflow.com/questions/64414944/hot-to-get-the-set-difference-of-two-2d-numpy-arrays-or-equivalent-of-np-setdif
    """
    nrows, ncols = A.shape
    dtype = dict(
        names=['f{}'.format(i) for i in range(ncols)],
        formats=ncols * [A.dtype]
    )
    C = np.setdiff1d(A.copy().view(dtype), B.copy().view(dtype))
    return C


# ----------------------------Specific Image Processing Algorithms-----------------------------


def find_np3d_from_bs(mask: npt.NDArray[np.uint8]) -> list[npt.NDArray[np.int64]]:
    """Find sparse representation of the contour locations for a contour mask.

    "bs"=binary segmentation mask

    The input lbl_im is an image with pixel values 0-N, typically returned from scipy.ndimage.label() or
    skimage.morphology.label().

    Args:
        mask: The binary mask to be segmented into contours represented in argwhere format

    Returns:
        A list of contours, each represented in a Mi * d ndarray of type np.int64. Each row is a location
        vector indicating the pixel location, and there are Mi pixels making up the ith contour.

        If some number < N in lbl_im corresponds to no pixels in the array, the corresponding entry will
        be None. This function will not label lbl_im == 0 (which is assumed to be the background class)
    """
    lbl_im = instance_label(mask)[0]
    return npindices_from_os(lbl_im)[0]


def npindices_from_os(
        lbl_im: npt.NDArray[np.int32],
        return_object_slices: bool = False,
        is_sparse: bool = False,
) -> tuple:
    """Find sparse representation of the contour locations for a contour mask.

    "os"=ordinal segmentation mask

    The input lbl_im is an image with pixel values 0-N, typically returned from scipy.ndimage.label() or
    skimage.morphology.label().

    Args:
        lbl_im: The image with pixel values 0-N, each number 1-N correspond to a separate object.
        return_object_slices: If True, return the boundbox used to find the corresponding mask image
        is_sparse: If True, input index labels may be large int but only a few indices are present; also
            return the corresponding indices after the other returning data

    Returns:
        A list of contours, each represented in a Mi * d ndarray of type np.int64. Each row is a location
        vector indicating the pixel location, and there are Mi pixels making up the ith contour.

        If some number in lbl_im corresponds to no pixels in the array, the corresponding entry will
        be None. This function will not label lbl_im == 0 (which is assumed to be the background class)

        The returned result list's index 0 correspond to the first non-background instance; if slices are
        returned, the ith index of result list will correspond to the ith index of slices list
    """
    lbl_ndim = lbl_im.ndim
    if is_sparse:
        unique, unique_inverse = np_unique(lbl_im, return_inverse=True)
        unique = unique[1:]  # remove 0 = background class
        lbl_im = unique_inverse

    object_slices = find_objects(lbl_im)
    result = []
    result_slices = []
    for slices in object_slices:
        if slices is None:
            result.append(None)
            continue

        i = len(result) + 1
        mask_argwhere = np.argwhere(lbl_im[slices] == i)
        mask_argwhere += np.array(tuple(s.start for s in slices), dtype=np.int64)
        result.append(mask_argwhere)
        result_slices.append(slices)
        assert mask_argwhere.shape[1] == lbl_ndim, (f'Expected row length to match label ndim, got shape '
                                                    f'{mask_argwhere.shape} and ndim={lbl_ndim}')

    ret = (result,)
    if return_object_slices:
        ret += (result_slices,)
    if is_sparse:
        assert len(result) == len(unique), f'Expected equal length, got {len(result)}, {len(unique)}'
        ret += (unique,)
    return ret


# ------------------------------------------Watershed------------------------------------------


def watershed(seg_bin, dist_thres=1., remove_smaller_than=None):
    """Run Watershed algorithm to perform instance segmentation.

    The result is an index labeled int64 mask

    Args:
        seg_bin: The binary [0, 1] 3d mask to run watershed on to separate blobs into instances.
        dist_thres: Only pixels this much into the contours are retained; pixels on contours surface are removed
        remove_smaller_than: Contours smaller than this value are added as part of neighboring contours (once)

    Returns:
        The instance segmented int64 mask from 0 to N, where N is higher than the number of objects. Note
        this algorithm does not guarantee there are N objects found, since some of the contours between 1
        and N will be removed as a filter step
    """
    # reference: https://docs.opencv.org/4.x/d3/db4/tutorial_py_watershed.html
    # fp_width = 2
    # fp = [(np.ones((fp_width, 1, 1)), 1), (np.ones((1, fp_width, 1)), 1), (np.ones((1, 1, fp_width)), 1)]
    # sure_bg = morph.binary_dilation(seg_bin, fp)
    sure_bg = seg_bin
    # sure_fg = morph.binary_erosion(seg_bin, fp)
    dist_transform = distance_transform_edt(seg_bin)
    sure_fg = dist_transform >= dist_thres
    unknown = sure_bg ^ sure_fg
    lbl_im = skimage.morphology.label(sure_fg, connectivity=1)

    # so that sure_bg is 1 and unknown region is 0
    lbl_im += 1
    lbl_im[unknown == 1] = 0
    result = skimage.segmentation.watershed(-dist_transform, lbl_im, connectivity=1)
    result -= result > 0  # we don't need to mark sure_bg as 1, make all marker id smaller by 1

    if remove_smaller_than is not None:
        # watershed again over the small object regions
        small_mask, big_mask = split_labeled_objects(result, remove_smaller_than, connectivity=1)
        lbl_im[small_mask] = 0
        result = skimage.segmentation.watershed(-dist_transform, lbl_im, connectivity=1)

        # at this point, there may be unfilled space, which are rare cases where lots of small objects make up a
        # larger connected part; these space should be cells but the individual objects do not meet size requirement
        unfilled = skimage.morphology.label(result == 0, connectivity=1)
        result += (unfilled != 0) * (unfilled + result.max())

        result -= result > 0  # we don't need to mark sure_bg as 1, make all marker id smaller by 1

    return result


def split_labeled_objects(lbl_im, size_thres, connectivity=1):
    component_sizes = np.bincount(lbl_im.ravel())
    small_inds = component_sizes < size_thres
    small_inds[0] = False
    small_mask = small_inds[lbl_im]
    big_inds = component_sizes >= size_thres
    big_inds[0] = False
    big_mask = big_inds[lbl_im]
    return small_mask, big_mask


def split_objects(seg, size_thres, connectivity=1):
    lbl_im = skimage.morphology.label(seg, connectivity=connectivity)
    return split_labeled_objects(lbl_im, size_thres, connectivity)


def round_object_detection_3sizes(seg, size_thres, dist_thres, rst, size_thres2, dist_thres2, rst2,
                                  remap_indices: bool = True):
    """detect round objects in a binary image

    Args:
        seg: the binary mask where we want to detect on
        size_thres: the size threshold below which size, contours are all kept as is
        dist_thres: threshold for seeding in the watershed algorithm
        rst:
        size_thres2:
        dist_thres2:
        rst2:
        remap_indices: If True, make sure `lbl_im.max()` is the number of objects; if this is False,
            `lbl_max()` may be larger than the actual number of objects detected

    Returns:
        Same shape image labeled 0, 1, ..., n, where 0 is background and 1...n are detected objects; note if
        remap_indices is False, there may not be n objects as some of the objects between 1 and n are removed
        from the list during filter step.
    """
    # objects too small cannot be connected. We use this property to first find small objects that must be single cells
    small_mask, big_mask = split_objects(seg, size_thres, connectivity=1)
    big_mask, big_mask2 = split_objects(big_mask, size_thres2, connectivity=1)

    # big objects may be a single large cell or overlapping small cells, run watershed to separate overlapping cells
    lbl_im = watershed(big_mask, dist_thres=dist_thres, remove_smaller_than=rst)
    lbl_im2 = watershed(big_mask2, dist_thres=dist_thres2, remove_smaller_than=rst2)

    # finally, combine the two set of cells
    small_labeled = skimage.morphology.label(small_mask, connectivity=1)
    lbl_im += (small_labeled != 0) * (small_labeled + lbl_im.max())
    lbl_im += (lbl_im2 != 0) * (lbl_im2 + lbl_im.max())
    # lbl_im = (lbl_im2 > 0) * 1 + (lbl_im > 0) * 2 + small_mask * 3

    if remap_indices:
        _, lbl_im = np_unique(lbl_im, return_inverse=True)

    return lbl_im


def voronoi_ndarray(im_shape: tuple, centroids: npt.NDArray[np.int32]) -> npt.NDArray[np.int32]:
    """Given k centroids in a voxel ndarray, color the array by voronoi algorithm

    Args:
        im_shape: Shape of the array to be colored
        centroids: The centroids in the space

    Returns:
        The indices of type np.int32 according to centroid order
    """
    import numpy as np
    from scipy.spatial import cKDTree

    # Algorithm provided by chatgpt-4o

    indices = np.indices(im_shape)
    grid_points = np.stack([idx.ravel() for idx in indices], axis=-1)

    # Use cKDTree to find the nearest centroid for each voxel
    tree = cKDTree(centroids)
    _, labels = tree.query(grid_points)
    return labels.astype(np.int32).reshape(im_shape)


# --------------------------------Statistical Analysis-------------------------------------


def Stats_MAE(counted, gt):
    """
    Params
        counted (list) - the counted cells in each image
        gt (Iterable[float]) - the ground truth number of cells in each image
    Returns
        the mean absolute difference between counted and gt
    """
    return np.abs(np.array(counted, dtype=np.float64) - np.array(gt, dtype=np.float64)).mean().item()


def Stats_ShowScatterPairComparisons(counted: np.array, gt, enumType: Type[enum.Enum]) -> None:
    """Show comparison between different algorithms counting results to the gt

    Args:
        counted: (a NCounter * NImages np.ndarray) The counted cells in each image, by each counter
        gt: (Iterable[float]) The ground truth number of cells in each image
        enumType: The enum class defined for counting
    """
    import matplotlib.pyplot as plt
    counting_method_inverse_dict = {item.value: item.name for item in enumType}
    nrow = int((len(counting_method_inverse_dict) - 1) / 6) + 1
    fig, axes = plt.subplots(nrow, 6, figsize=(24, nrow * 4), sharex=True, sharey=True)

    gt_arr = np.array(gt, dtype=np.float64)
    for i in range(counted.shape[0]):
        X, Y = gt_arr, counted[i]

        iax, jax = i // 6, i % 6
        ax: plt.Axes = axes[iax, jax]
        ax.set_box_aspect(1)
        ax.set_title(counting_method_inverse_dict[i])
        ax.scatter(X, Y)

    plt.show()


if __name__ == '__main__':
    centroids = ((2, 2), (0, 1), (0, 0))
    colored = voronoi_ndarray((3, 3), np.array(centroids, dtype=np.int32))
    assert colored.dtype == np.int32, f'got colored.dtype={colored.dtype}'
    colored[2, 0] = 0  # ambiguous as to whether this is 0 or 2, for testing purpose we don't mind this
    reference = np.array(
        ((2, 1, 1),
         (2, 1, 0),
         (0, 0, 0)), dtype=np.int32
    )
    assert (colored == reference).sum().item() == colored.size, f'got array {colored}'

    # lbl_im = np.array(
    #     (((1, 3),
    #       (0, 0)),
    #      ((0, 0),
    #       (0, 2))),
    #     dtype=np.int32
    # )
    # _contours_np3d, _ids = npindices_from_os(lbl_im, is_sparse=True)
    # for cnt in _contours_np3d:
    #     print('contour')
    #     print(cnt, cnt.shape)
    # print(_ids)
